# -*- coding:utf-8 -*-
# @Time: 2022/11/23 21:42
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: excel.py
import os
from typing import List

from MelodieInfra.jsonobject import StringProperty, JsonObject, DefaultProperty, ObjectProperty
from .base import ExcelMeta


def path_validator(path: str):
    if not os.path.exists(path):
        raise FileNotFoundError(path)


class ExcelWriteRequest(JsonObject):
    path = StringProperty(
        required=True, validators=path_validator, name="path")
    data = DefaultProperty(required=True, name='data')
    sheet = StringProperty(required=True, name="sheet")


class ExcelReadSheetRequest(JsonObject):
    path = StringProperty(name="path",
                          required=True, validators=path_validator)
    sheet = StringProperty(name="sheet", required=False)


class ExcelReadSheetResponse(JsonObject):
    payload = DefaultProperty(name="payload", required=True)
    meta = ObjectProperty(ExcelMeta, name="meta", required=True)

    @staticmethod
    def create(payload, sheet: str, all_sheets: List[str]):
        return ExcelReadSheetResponse(
            payload=payload,
            meta=ExcelMeta({
                "currentSheet": sheet,
                "sheetNames": all_sheets,
                "widget": "table",
                "type": "excel"
            })
        )



from typing import Dict, Any


class Element:
    def set_params(self, params: Dict[str, Any]):
        """

        :param params:
        :return:
        """
        for paramName, paramValue in params.items():
            assert (
                paramName in self.__dict__.keys()
            ), f"param named {paramName}, value {paramValue} not in Agent.params:{self.__dict__.keys()}"
            setattr(self, paramName, paramValue)



# -*- coding:utf-8 -*-
# @Time: 2022/12/8 22:47
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: actions.py
from typing import Callable, Dict, List, Optional
from MelodieInfra.jsonobject import JsonObject, StringProperty, ObjectProperty
from .vis_charts import JSONBase


class BasicOprandType(JSONBase):
    def __init__(self):
        self.type = ''

    def __eq__(self, other):
        assert isinstance(other, BasicOprandType)
        return other.type == self.type


class FileOprandType(BasicOprandType):
    """
    File Oprand.
    In JS, it is a `File()` Object.

    """

    def __init__(self):
        self.type = 'file'


class ResponseOprandType(BasicOprandType):
    """
    Response oprand.
    In JS, it is an axios response.

    """

    def __init__(self):
        self.type = 'response'


class Operation(JSONBase):

    def __init__(self, oprands: Optional[List['Operation']]):
        self.name = ""
        self.oprands: List[Operation] = oprands if oprands is not None else []
        assert isinstance(self.oprands, list)
        self.optypes = []
        self.rettype = None
        self.type = None

    def check_type(self):
        assert len(self.oprands) == len(self.optypes), (self.optypes, self.oprands)
        for oprand, optype in zip(self.oprands, self.optypes):
            assert oprand.rettype == optype, (oprand.rettype, optype)


class ResponseConversionOperation(Operation):
    pass


class ResponseToFile(ResponseConversionOperation):
    names_map = {"default_name": "defaultName"}

    def __init__(self, default_name: str, oprands: Optional[List['Operation']] = None):
        super().__init__(oprands)
        self.name = 'op-response-to-file'
        self.default_name = default_name
        self.optypes = []
        self.rettype = FileOprandType()


class DownloadOperation(Operation):

    def __init__(self, oprands: Optional[List['Operation']] = None):
        super().__init__(oprands)
        self.name = "op-download"
        self.optypes = [FileOprandType()]
        self.rettype = None
        print(self.oprands)


class Action(JSONBase):
    handlers_map = {}

    def __init__(self, key: str, text: str, operation: Operation, handler: Callable):
        self.key = key
        self.text = text
        self.operation = operation
        self.operation.check_type()
        self._handler = handler
        self.children = []
        Action.handlers_map[key] = handler

    def add_sub_action(self, action: "Action"):
        self.children.append(action)

    @classmethod
    def dispatch(cls, key, args: Dict = None):
        if args is None:
            args = {}
        return cls.handlers_map[key](**args)



# -*- coding:utf-8 -*-
# @Time: 2022/11/21 16:48
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: __init__.py.py

from .exceptions import *
from .pretty_warnings import show_prettified_warning, show_link
from .troubleshooting import OSTroubleShooter



import logging

from typing import Any, Callable, List, Dict, TYPE_CHECKING, Tuple, TypeVar, Union, Optional

from MelodieInfra import NUMERICAL_TYPE
from MelodieInfra.models.typeutils import REAL_NUM_TYPE

logger = logging.getLogger(__name__)


class JSONBase:
    names_map = {}

    def to_json(self):
        d = {}
        for k, v in self.__dict__.items():
            k = k if k not in self.__class__.names_map else self.__class__.names_map[k]
            if k.startswith("_"):
                continue
            elif hasattr(v, "to_json"):
                d[k] = v.to_json()
            elif isinstance(v, dict):
                new_dict = {}
                for new_k, new_v in v.items():
                    if hasattr(new_v, "to_json"):
                        new_dict[new_k] = new_v.to_json()
                    else:
                        new_dict[new_k] = new_v
                d[k] = new_dict
            elif isinstance(v, (list, tuple)):
                new_list = []
                for new_v in v:
                    if hasattr(new_v, "to_json"):
                        new_list.append(new_v.to_json())
                    else:
                        new_list.append(new_v)
                d[k] = new_list
            else:
                d[k] = v
        return d


class ChartBase:

    def reset(self):
        pass

    def set_data_source(self, *args, **kwargs) -> Any:
        """
        Set the data source of chart

        :param args:
        :param kwargs:
        :return:
        """
        return self

    def update(self):
        pass


class ChartSeries(JSONBase):
    def __init__(self, name: str, source: "Callable[[], int]" = None):
        self.seriesName = name
        self._source = source
        self.data: List[Union[int, float]] = []
        self.latest_data: Optional[Union[int, float]] = None

    def add_data_value(self, data_value: Union[int, float]):
        self.latest_data = data_value
        self.data.append(data_value)

    def reset(self):
        self.data = []
        self.latest_data = None

    def set_data_source(self, datasource: "Callable[[], Any]"):
        self._source = datasource

    def update(self, current_step: int):
        d = self._source()
        self.add_data_value(d)
        assert current_step == len(self.data), (current_step, len(self.data))


class Chart(JSONBase, ChartBase):
    def __init__(self):
        self._index = {}
        self.type = "line"
        self.series: Optional[List[ChartSeries]] = None

    def set_series(self, series_names: List[str]):
        self._index = {series_name: i for i, series_name in enumerate(series_names)}
        self.series: List[ChartSeries] = [ChartSeries(name) for name in series_names]
        return self

    def set_data_source(self, data_source: Dict[str, Callable[[], REAL_NUM_TYPE]]):
        """
        Set data source of each series

        :param data_source:
        :return:
        """
        self.set_series([k for k in data_source.keys()])
        for series_name in data_source.keys():
            # series = self.get_series(series_name)
            self.set_series_data_source(series_name, data_source[series_name])
        return self

    def get_series(self, series_name: str) -> ChartSeries:
        return self.series[self._index[series_name]]

    def get_series_by_index(self, series_index: int) -> ChartSeries:
        return self.series[series_index]

    def reset(self):
        for series in self.series:
            series.reset()

    def update(self, current_step: int):
        for series in self.series:
            series.update(current_step)

    def get_series_data(self):
        chart_series_data = []
        for i in range(len(self.series)):
            s = {
                "name": self.get_series_by_index(i).seriesName,
                "value": self.get_series(
                    self.get_series_by_index(i).seriesName
                ).latest_data,
            }
            chart_series_data.append(s)
        return chart_series_data

    def set_series_data_source(self, series_name: str, datasource):
        """
        Set data source for one series

        :param series_name:
        :param datasource:
        :return:
        """
        self.get_series(series_name).set_data_source(datasource)
        return self


class CandleStickChart(Chart):

    def __init__(self):
        self.type = "candlestick"
        self._series_name = 'series'
        self.series: Tuple[ChartSeries] = (ChartSeries(self._series_name),)

    def get_series(self, series_name: str) -> ChartSeries:
        if series_name == self._series_name:
            return self.series[0]
        else:
            raise ValueError(series_name)

    def set_data_source(self, datasource: Callable[
        [], Tuple[REAL_NUM_TYPE, REAL_NUM_TYPE, REAL_NUM_TYPE, REAL_NUM_TYPE]]) -> 'CandleStickChart':
        """

        :param datasource: Callable, should return a tuple with four real number -- (open, close, low, high)
        :return:
        """
        self.get_series('series').set_data_source(datasource)
        return self


class Gauge(JSONBase, ChartBase):
    def __init__(self, source: Callable[[], Union[int, float, Dict]]):
        self._sources: Dict[str, Callable[[], Union[int, float]]] = {}
        self.value: Dict[str, Union[int, float]] = {}

    def reset(self):
        pass

    def update(self, current_step):
        self.value = self._source()


class PieChart(JSONBase, ChartBase):
    def __init__(self):
        self.type = "pie"
        self._sources: Dict[str, Callable[[], Union[int, float]]] = {}
        self._source: Optional[Callable[[], Dict[Union[int, float]]]] = None
        self.value: Dict[str, Union[int, float]] = {}
        self._mode = 'single'

    def add_variable(self, variable_name: str, source: Callable[[], Union[int, float]]):
        self.value[variable_name] = 0
        self._sources[variable_name] = source

    def update(self, current_step):
        assert isinstance(self.value, dict)
        if self._mode == 'single':
            for varname in self._sources.keys():
                self.value[varname] = self._sources[varname]()
        else:
            self.value = self._source()

    def reset(self):
        self.value = {}

    def to_json(self):
        return {'type': self.type, 'series': self.get_series_data()}

    def get_series_data(self):
        return [{"name": k, 'value': v} for k, v in self.value.items()]


class BarChart(PieChart):
    def __init__(self):
        super(BarChart, self).__init__()
        self.type = 'bar'
        self._mode = 'single'

    def set_data_source(self, data_source: Dict[str, Callable[[], Union[int, float]]]):
        self._sources = data_source
        return self


class ChartManager(JSONBase):
    def __init__(self):
        self.charts: Dict[str, Union[Chart, Gauge]] = {}

    def add_chart(self, chart_name: str, chart):
        self.charts[chart_name] = chart
        return chart

    def add_line_chart(self, chart_name: str) -> Chart:
        return self.add_chart(chart_name, Chart())

    def add_piechart(self, chart_name: str) -> PieChart:
        return self.add_chart(chart_name, PieChart())

    def add_barchart(self, chart_name: str) -> BarChart:
        return self.add_chart(chart_name, BarChart())

    def add_candlestick_chart(self, chart_name: str) -> CandleStickChart:
        return self.add_chart(chart_name, CandleStickChart())

    def all_chart_names(self):
        return set(self.charts.keys())

    def get_chart(self, chart_name: str) -> Chart:
        return self.charts[chart_name]

    def get_current_data(self) -> List[Dict[str, List]]:
        current_data = []
        for chart_name in self.all_chart_names():
            chart = self.get_chart(chart_name)
            # chart_series_data = []
            # for i in range(len(chart.series)):
            #     s = {
            #         "name": chart.get_series_by_index(i).seriesName,
            #         "value": chart.get_series(
            #             chart.get_series_by_index(i).seriesName
            #         ).latest_data,
            #     }
            #     chart_series_data.append(s)
            current_data.append({"chartName": chart_name, "series": chart.get_series_data()})
        # print("current_data", current_data)
        return current_data

    def reset(self):
        for chart_name, chart in self.charts.items():
            chart.reset()
        logger.info("Chart manager reset!")

    def update(self, current_step):
        for chart_name, chart in self.charts.items():
            chart.update(current_step)



# cython:language_level=3
# -*- coding:utf-8 -*-
# @Time: 2021/10/3 20:58
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: grid.pyx

from typing import List, TypeVar

T = TypeVar("T")

def sample(population: List[T], sample_num: int) -> List[T]: ...



import logging
import os
from typing import Union, Dict, TYPE_CHECKING, Type, Optional, List, Tuple

import pandas as pd
import sqlalchemy
from sqlalchemy.exc import OperationalError

from ..exceptions import MelodieExceptions

if TYPE_CHECKING:
    from MelodieInfra.config.config import Config

TABLE_DTYPES = Dict[
    str,
    Union[
        str, Type[str], Type[float], Type[int], Type[complex], Type[bool], Type[object]
    ],
]

SQLITE_FILE_SUFFIX = ".sqlite"

logger = logging.getLogger(__name__)


class DBConn:
    """
    DBConn provides API to write to/read from the database.

    """
    table_dtypes: Dict[str, TABLE_DTYPES] = {}
    SCENARIO_TABLE = "simulator_scenarios"
    ENVIRONMENT_RESULT_TABLE = "environment_result"

    def __init__(
            self, db_name: str, db_type: str = "sqlite", conn_params: Dict[str, str] = None
    ):
        self.db_name = db_name

        if db_type not in {"sqlite"}:
            MelodieExceptions.Data.InvalidDatabaseType(db_type, {"sqlite"})
        if db_type == "sqlite":
            if conn_params is None:
                conn_params = {"db_path": ""}
            elif not isinstance(conn_params, dict):
                raise NotImplementedError
            elif conn_params.get("db_path") is None:
                conn_params["db_path"] = ""

            self.db_path = conn_params["db_path"]
            self.connection = self.create_connection(db_name)
        else:
            raise NotImplementedError

    def get_engine(self):
        """
        Get the connection

        :return: Database engine
        """
        return self.connection

    def create_connection(self, database_name) -> sqlalchemy.engine.Engine:
        """
        Create a connection to the sqlite database.

        :param database_name: Name of sqlite database file.
        :return:
        """
        return sqlalchemy.create_engine(
            f"sqlite:///{os.path.join(self.db_path, database_name + SQLITE_FILE_SUFFIX)}"
        )

    @classmethod
    def create_from_existing_db(self, type: str, meta: Dict):
        """
        Create from existing database

        :param type: Only support 'sqlite' now.
        :param meta: A dict like ``{'path': 'path-to-sqlite'}``
        :return: Database engine.
        """
        assert type in {"sqlite"}
        return sqlalchemy.create_engine(f"sqlite:///{os.path.join(meta['path'])}")

    @classmethod
    def register_dtypes(cls, table_name: str, dtypes: TABLE_DTYPES):
        """
        Register data types of a table for sqlalchemy.
        """
        MelodieExceptions.Assertions.Type("dtypes", dtypes, dict)
        if table_name in cls.table_dtypes:
            raise ValueError(
                f"Table dtypes of '{table_name}' has been already defined!"
            )
        cls.table_dtypes[table_name] = dtypes

    @classmethod
    def get_table_dtypes(cls, table_name: str) -> TABLE_DTYPES:
        """
        Get the data type of a table.
        If table data type is not specified, return an empty dict.

        :param table_name: Name of table in database.
        :return:
        """
        if table_name in cls.table_dtypes:
            return cls.table_dtypes[table_name]
        else:
            return {}

    def close(self):
        """
        Close DB connection.
        """
        self.connection.dispose()

    def clear_database(self):
        """
        Clear the database, deleting all tables.
        """
        logger.info(f"Database contains tables: {self.connection.table_names()}.")
        table_names = list(self.connection.table_names())
        for table_name in table_names:
            self.connection.execute(f"drop table {table_name}")
        logger.info(f"Database drops tables: {table_names}.")

    def write_dataframe(
            self,
            table_name: str,
            data_frame: pd.DataFrame,
            data_types: Optional[TABLE_DTYPES] = None,
            if_exists="append",
    ):
        """
        Write a dataframe to database.

        :param table_name: Name of table.
        :param data_frame: The dataframe to be written into the database.
        :param data_types: The data type for columns.
        :param if_exists: A string in {'replace', 'fail', 'append'}.
        :return:
        """
        if data_types is None:
            data_types = DBConn.get_table_dtypes(table_name)
        logger.debug(f"datatype of table `{table_name}` is: {data_types}")
        data_frame.to_sql(
            table_name,
            self.connection,
            index=False,
            dtype=data_types,
            if_exists=if_exists,
        )

    def read_dataframe(
            self,
            table_name: str,
            id_scenario: Optional[int] = None,
            id_run: Optional[int] = None,
            conditions: List[Tuple[str, str]] = None,
    ) -> pd.DataFrame:
        """
        Read a table and return all content as a dataframe.

        For example:

        .. code-block:: python
            :linenos:

            df = create_db_conn(config).read_dataframe('agent_params', id_scenario=0,
                                                       conditions=[('id', "<=100"), ("health_state", '=1')])
            print(df)

        :param table_name: Name of table inside database.
        :param id_scenario: Filter of scenario
        :param id_run: Filter of run_id
        :param conditions: Custom conditions
        """
        where_condition_phrase = ""
        condition_phrases = []
        if conditions is not None:
            condition_phrases.extend([item[0] + item[1] for item in conditions])
        if id_scenario is not None:
            # assert isinstance(id_scenario, int)
            condition_phrases.append(f"id_scenario={id_scenario}")
        if id_run is not None:
            condition_phrases.append(f"id_run={id_scenario}")
        try:
            sql = f"select * from {table_name}"
            if len(condition_phrases) != 0:
                sql += " where " + " and ".join(condition_phrases)
            logger.debug("Querying database: " + sql)
            return pd.read_sql(sql, self.connection)
        except OperationalError:
            import traceback

            traceback.print_exc()
            raise MelodieExceptions.Data.AttemptingReadingFromUnexistedTable(table_name)

    def drop_table(self, table_name: str):
        """
        Drop table if it exists.

        :param table_name: The name of table to drop.
        :return:
        """
        self.connection.execute(f"drop table if exists  {table_name} ;")

    def query(self, sql) -> pd.DataFrame:
        """
        Execute sql command and return the result by pd.DataFrame.

        :param sql: SQL phrase to execute.
        :return:
        """
        return pd.read_sql(sql, self.connection)


def create_db_conn(config: "Config") -> DBConn:
    """
    create a Database by current config

    :return: DBConn object.
    """

    return DBConn(config.project_name, conn_params={"db_path": config.output_folder})


def get_sqlite_filename(config: "Config") -> str:
    """
    Get SQLite database filename from Melodie project config.
    """
    return os.path.join(config.output_folder, config.project_name + SQLITE_FILE_SUFFIX)



import copy
import logging
import pickle
from typing import List, Optional, Union, TYPE_CHECKING

import numpy as np
import pandas as pd

from MelodieInfra import MelodieExceptions
from .element import Element

if TYPE_CHECKING:
    from Melodie import Calibrator, Simulator
    from .data_loader import DataFrameInfo, MatrixInfo
logger = logging.getLogger(__name__)


class Scenario(Element):
    """
    Scenario contains a set of parameters used in simulation model.
    It is created before the initialization of ``Model``.

    """
    def __init__(self, id_scenario: Optional[Union[int, str]] = None):
        """
        :param id_scenario: the id of scenario. if None, this will be self-increment from 0 to scenarios_number-1
        """
        super().__init__()
        if (id_scenario is not None) and (not isinstance(id_scenario, (int, str))):
            raise MelodieExceptions.Scenario.ScenarioIDTypeError(id_scenario)
        self._parameters = []
        self.manager: Union["Calibrator", "Simulator", None] = None
        self.id = id_scenario
        self.run_num = 1
        self.period_num = 0

    def copy(self) -> "Scenario":
        """
        Copy current scenario to a new scenario.

        :return: New scenario object.
        """
        new_scenario = self.__class__()
        for property_name, property in self.__dict__.items():
            # assert property_name in new_scenario.__dict__, (property_name, new_scenario.__dict__)
            setattr(new_scenario, property_name, property)
        for parameter in self._parameters:
            parameter.init = getattr(self, parameter.name)
        return new_scenario

    def setup(self):
        """
        Setup method, be sure to inherit it on the custom scenario class.
        """
        pass

    def to_dict(self):
        """
        Convert this scenario object to a dict.

        :return: A ``dict``, ``property_name->property_value``
        """
        d = {}
        for k in self.__dict__.keys():
            v = self.__dict__[k]
            d[k] = v
        return d

    def to_json(self):
        """
        Convert this scenario to a dict without concerning non-serializable properties.

        :return: a ``dict``, ``property_name->property_value``, without non-serializable properties
        """
        d = {}
        for k in self.__dict__.keys():
            if k not in {"manager"}:
                d[k] = self.__dict__[k]
        return d

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.__dict__}>"

    def get_dataframe(self, df_info: "DataFrameInfo") -> pd.DataFrame:
        """
        Get dataframe from scenario

        :param df_info:
        :return: pandas dataframe.
        """
        assert self.manager is not None
        return self.manager.get_dataframe(df_info.df_name)

    def get_matrix(self, matrix_info: "MatrixInfo") -> np.ndarray:
        """
        Get matrix from scenario.

        :param matrix_info:
        :return: 2D numpy array
        """
        assert self.manager is not None
        return self.manager.get_matrix(matrix_info.mat_name)



"""
This script file is only concerned about the system information, and generate the report.

Pasting the generated report onto GitHub issues can be of great convenience.
"""
import os.path
import platform


def is_windows() -> bool:
    return platform.system().lower().find("windows") != -1


def melodie_version() -> str:
    """
    Get the version of Melodie.
    :return:
    """
    try:
        with open(
            os.path.join(os.path.dirname(os.path.dirname(__file__)), "version.txt")
        ) as f:
            return f.read()
    except FileNotFoundError:
        import traceback

        traceback.print_exc()
        return ""


def get_system_info(print_info=True) -> str:
    """
    Get the system info
    :param print_info: Whether the information be directly printed to your console.
    :return:
    """
    import numpy
    import Cython

    info = f"""
    Here is the information for Melodie
    ===================================
    Melodie Version  :  {melodie_version()}
    Python Version   :  {platform.python_version()}
    Python Arch      :  {platform.architecture()}
    Platform Detail  :  {platform.platform()}
    Cython Version   :  {Cython.__version__}
    Numpy Version    :  {numpy.__version__}
    """
    if print_info:
        print(info)
    return info



# -*- coding:utf-8 -*-
# @Time: 2022/11/24 23:01
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: db.py
import json
from typing import List

import pandas as pd
import sqlalchemy
from sqlalchemy.exc import OperationalError
from MelodieInfra.models import DatabaseQueryRequest, DataServiceStatus, DataServiceState, DatabaseBasicRequest


def get_table_names(conn_string: str) -> List[str]:
    """

    :return:
    """
    conn = sqlalchemy.create_engine(conn_string)
    table_names: List[str] = conn.table_names()
    return table_names


class DatabaseService():
    @staticmethod
    def query_database(req: DatabaseQueryRequest) -> DataServiceStatus:
        try:
            return DataServiceStatus(
                status=DataServiceState.SUCCESS,
                msg="Query Succeeded!",
                data=json.loads(DatabaseService.execute_sql(req.connection_string, req.sql).to_json(
                    orient="table", indent=4, index=False)),
            )
        except OperationalError as e:
            import traceback
            traceback.print_exc()
            return DataServiceStatus(
                status=DataServiceState.ERROR,
                msg=f"{e}",
                data=None
            )

    @staticmethod
    def table_names(req: DatabaseBasicRequest):
        try:
            return DataServiceStatus(
                status=DataServiceState.SUCCESS,
                msg="Get table names succeeded!",
                data=get_table_names(req.connection_string)
            )
        except OperationalError as e:
            import traceback
            traceback.print_exc()
            return DataServiceStatus(
                status=DataServiceState.ERROR,
                msg=f"{e}",
                data=None
            )

    @staticmethod
    def execute_sql(conn_string, sql: str):
        engine = sqlalchemy.create_engine(conn_string)
        return pd.read_sql(sql, engine)



# -*- coding:utf-8 -*-
# @Time: 2022/11/10 14:36
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: unsafe.py

class Unsafe:
    @staticmethod
    def create_getter_by_str(rvalue_expr):
        func_str = f"""
def getter(scenario):
    return {rvalue_expr}
        """
        local_vars = {}
        exec(func_str, {}, local_vars)
        return local_vars['getter']

    @staticmethod
    def create_setter_by_str(lvalue_expr):
        func_str = f"""
def setter(scenario, value):
    {lvalue_expr} = value
        """
        local_vars = {}
        exec(func_str, {}, local_vars)
        return local_vars['setter']



from collections import Counter
from typing import TYPE_CHECKING, Any, List, Callable, Dict, Set

if TYPE_CHECKING:
    from Melodie import Agent


def assert_exc_occurs(exc_id: int, func: Callable):
    """
    Assert that this exception will occur.
    """
    try:
        func()
        assert False
    except MelodieException as e:
        import traceback
        traceback.print_exc()
        assert e.id == exc_id


def assert_exc_type_occurs(exc_type: BaseException, func: Callable):
    """
    Assert that exception of `exc_type` must occur.

    """
    try:
        func()
        assert False
    except BaseException as e:
        import traceback
        traceback.print_exc()
        assert isinstance(e, exc_type)


class MelodieException(Exception):
    def __init__(self, exc_id: int, text: str):
        text = f"{text} <Error ID {exc_id}>"
        super(MelodieException, self).__init__(text)
        self.id = exc_id


class MelodieExceptions:
    class Assertions:
        @staticmethod
        def Type(name, obj, expected_type):
            """
            Assert that the type of `obj` is the `expected_type`.
            """
            if not isinstance(obj, expected_type):
                raise MelodieExceptions.General.TypeError(name, obj, expected_type)

        @staticmethod
        def IsNone(name, obj):
            """
            Assert variable `obj` is None.

            """
            if obj is not None:
                raise TypeError(f"{name} should be None, however it is not None.")

        @staticmethod
        def NotNone(name, obj):
            """
            Assert that the `obj` is not None
            """
            if obj is None:
                raise TypeError(f"{name} should not be None, however it is None.")

    class General:
        @staticmethod
        def TypeError(name, obj, expected_type: type):
            """
            The encapsulation of TypeError
            """
            return TypeError(
                f"{name} should be a {expected_type}, however it was {type(obj)}, value {obj}"
            )

        @staticmethod
        def NoAttributeError(obj, attr_name: str):
            """
            The encapsulation of AttributeError
            """
            return AttributeError(f"Object {obj} has no attribute {attr_name}")

    class Program:
        """
        Errors related to programming

        * code: 1000
        """

        ID = 1000

        class Variable:
            ID = 1010

            @staticmethod
            def VariableInvalid(var_desc: str, var_value: Any, expected_value: Any):
                """
                If this occurs, please check the value of variable

                * code: 1011
                """
                return MelodieException(
                    1011,
                    f"Variable {var_desc} is {var_value}, but expected {expected_value} ",
                )

            @staticmethod
            def VariableNotInSet(var_desc: str, var_value: Any, allowed_set: Set[Any]):
                """
                Please check if variable is in the allowed set.

                * code: 1012
                * example: If the allowed set is {"apple", "pears"}, but the variable value is "banana"
                """
                return MelodieException(
                    1012,
                    f"Variable {var_desc} is {var_value}, not in allowed set {allowed_set} ",
                )

        class Function:
            ID = 1020

            @staticmethod
            def FunctionArgsNumError(
                    func: Callable, expected_arg_num: int, actual_arg_num: int
            ):
                """
                Function should have correct number of arguments. If not, this error will be raised.

                * code: 1021
                * example: Expecting 1 argument, but the function was ``lambda a, b: a+b``
                """
                return MelodieException(
                    1021,
                    f"There should be {expected_arg_num} for function {func}, "
                    f"but the actual argument number was {actual_arg_num}",
                )

    class State:
        ID = 1100

        @staticmethod
        def StateNotFoundError(state, all_states):
            """ """
            return MelodieException(
                1101,
                f"State {repr(state)} is not defined. All states are: {all_states}",
            )

        @staticmethod
        def CannotMoveToNewStateError(
                old_state, new_state, all_possible_new_states: set
        ):
            if len(list(all_possible_new_states)) == 0:
                return MelodieException(
                    1102,
                    f"Current state is {repr(old_state)}, on which the status could only move to"
                    f" itself. However the new state was {repr(new_state)}",
                )
            else:
                return MelodieException(
                    1102,
                    f"Current state is {repr(old_state)}, on which the status could only move to"
                    f" {all_possible_new_states}. However the new state was {repr(new_state)}",
                )

        @staticmethod
        def NotAStateAttributeError(agent_cls, state_attr: str):
            return MelodieException(
                1103, f"Class {agent_cls} has not defined state attribute {state_attr}"
            )

    class Scenario:
        ID = 1200

        @staticmethod
        def ScenarioIDDuplicatedError(id_scenario):
            """
            Scenario ID should be unique. If there are identical scenario-id in the scenario table or generated,
            this error will be raised.

            * code: 1201
            * example: The scenario table below could make this error occur.

            .. list-table:: Scenarios
               :widths: 25 25 25 25
               :header-rows: 1

               * - ID
                 - env_param_1
                 - env_param_2
                 - ...
               * - 0
                 - 10
                 - 20
                 - ...
               * - 0
                 - 15
                 - 20
                 - ...

            In this table, the ids of scenarios are identical, which will cause this exception.

            """
            return MelodieException(
                1201, f"Scenario id {id_scenario} was duplicated, which is not allowed."
            )

        @staticmethod
        def ScenarioIDTypeError(id_scenario):
            """
            Scenario ID should be integer type.

            * code: 1202
            * hint: Please check carefully to make sure the scenario id is integer.
            """
            return MelodieException(
                1202,
                f"Scenario id {id_scenario} should be int. However its type was {type(id_scenario)}.",
            )

        @staticmethod
        def NoValidScenarioGenerated(scenarios):
            """
            Operators (Simulator/Trainer/Calibrator) will generate a list of scenarios.
            If no scenario has been generated, this error will be raised.

            * code: 1204

            Hint:

            * If you overrode the ``generate_scenarios`` method of DataframeLoader, make sure valid value returned.
            * If you are using ``xlsx`` tables, make sure the first page of ``<operator_type>_scenarios.xlsx``
              is not empty.

            """
            return MelodieException(
                1204,
                f"The scenario manager has not generated any valid scenarios. "
                f"The scenarios generated by gen_scenarios() was {scenarios},"
                f"please make sure gen_scenarios() returns a list of Scenario.",
            )

        @staticmethod
        def ParameterRedefinedError(parameter_name: str, all_params: List):
            """
            For the interactive parameters of Scenario, the name should be unique.
            If two parameters has the same name, this exception will be raised.

            * code: 1205
            """
            return MelodieException(
                1205,
                f'A parameter with same name "{parameter_name}" already existed! all parameters are: {all_params}',
            )

    class Agents:
        ID = 1300

        @staticmethod
        def AgentListEmpty(agent_manager):
            """
            * code: 1301
            """
            return MelodieException(
                1301, f"Agent manager {agent_manager} contains no agents!"
            )

        @staticmethod
        def AgentPropertyNameNotExist(property_name, agent):
            """
            * code: 1302
            """
            return MelodieException(
                1302,
                f"Agent {agent} does not have property {property_name}.",
            )

        @staticmethod
        def AgentIDConflict(agent_container_name: str, agent_ids: List[int]):
            """
            Agent ID should be unique.

            * code: 1303
            """
            c = Counter(agent_ids)
            duplicated_ids = [
                agent_id for agent_id, times in c.most_common() if times > 1
            ]

            return MelodieException(
                1303,
                f"Agent container `{agent_container_name}` has duplicated agent IDs: {duplicated_ids}.",
            )

    class Environment:
        ID = 1400

    class Data:
        """
        This class is used when external data is imported or exported.
        """

        ID = 1500

        @staticmethod
        def TableNameAlreadyExists(table_name: str, existed: str):
            return MelodieException(
                1501,
                f"Table Named {table_name} does not exist. All existed tables are: {existed}",
            )

        @staticmethod
        def StaticTableNotRegistered(table_name: str, all_table_names: List[str]):
            """
            Static table should be registered before use.

            * code: 1502
            * hint: Make sure this table is correctly registered in the DataLoader.
            """
            return MelodieException(
                1502,
                f"Table '{table_name}' is not registered. All registered tables are: {all_table_names}.",
            )

        @staticmethod
        def AttemptingReadingFromUnexistedTable(table_name):
            """
            When reading the database, the table must exist in the database.
            * code: 1503
            """
            return MelodieException(1503, f"Table '{table_name}' does not in database.")

        @staticmethod
        def ObjectPropertyTypeUnMatchTheDataFrameError(
                param_name: str,
                param_type: type,
                dataframe_dtypes: Dict[str, type],
                agent: "Agent",
        ):
            """
            Object type should match the type defined in dataframe.

            * code: 1504
            * example: If the dataframe has type float, and agent property was str.
            """
            return MelodieException(
                1504,
                f"The Agent property '{param_name}' is of type {param_type}, but the corresponding column "
                f"of the dataframe is of type {dataframe_dtypes[param_name]}.\n"
                f"The agent that offended is: {agent}",
            )

        @staticmethod
        def TableNameInvalid(table_name):
            """
            The table name should be an identifier.

            * code: 1505
            * hint: Do not use "+", "-", "*", "/" or other special characters.
            * example: ``table-123``; ``table+123``
            """
            return MelodieException(
                1505,
                f"Table name '{table_name}' is invalid."
                f"The expected table name should be an identifier.",
            )

        @staticmethod
        def TableNotFound(table_name: str, all_tables: dict):
            """
            Table may not be registered.

            * code: 1506

            """
            return MelodieException(
                1506,
                f"Table '{table_name}' is not found. All registered tables are: {set(all_tables.keys())}",
            )

        @staticmethod
        def InvalidDatabaseType(database: str, supported_db_types: Set[str]):
            """
            * code: 1508
            """
            return MelodieException(
                1508,
                f"Melodie only support these kinds of databases: {supported_db_types}.\n"
                f"Database type {database} is not supported right now.\n"
                "Is there any spelling problems with the database name?\n"
                "If you do not need to write data into database, please pass `None` as DB type.",
            )

        @staticmethod
        def NoDataframeLoaderDefined():
            """
            Dataframe Loader must be defined if you want to use the static table.

            * code: 1509
            """
            return MelodieException(
                1509,
                f"No dataframe loader defined for the Simulator/Calibrator/Trainer.",
            )

        @staticmethod
        def ColumnNameConsistencyError(
                df_name: str, missing: Set[str], undefined: Set[str]
        ):
            """
            The column name in code should be consistent to the table.

            * code: 1510
            """
            return MelodieException(
                1510,
                f"Column name inconsistent: '{df_name}'\n"
                + ("" if len(missing) == 0 else f"Missing columns: {missing}\n")
                + ("" if len(undefined) == 0 else f"Undefined columns: {undefined}"),
            )

    class Tools:
        """
        This class is for errors related to dev tools such as MelodieStudio
        """

        ID = 1600

        @staticmethod
        def MelodieStudioUnAvailable():
            """
            Melodie studio must be started before visualizer process start.

            * code: 1601
            """
            return MelodieException(
                1601,
                f"Connection to Melodie Studio was refused. It seems that Melodie studio is not "
                f"started yet. Please start Melodie Studio.",
            )

    class Visualizer:
        """
        This class is for errors related to visualizer.
        """

        ID = 1700

        class Charts:
            ID = 1700

            @staticmethod
            def ChartNameAlreadyDefined(chart_name: str, all_chart_names: List[str]):
                """
                * code: 1701
                """
                return MelodieException(
                    1701,
                    f"Chart name '{chart_name}' is already defined. All chart names are: {all_chart_names}",
                )



# cython:language_level=3
from .basics cimport Agent
from .agent_list cimport AgentList
cimport numpy as np
from libcpp.vector cimport vector
from libcpp.unordered_set cimport unordered_set as cpp_set
from libcpp.map cimport map as cpp_map
from libcpp.string cimport string as cpp_string

cdef class GridItem(Agent):
    cdef public long x
    cdef public long y
    cdef public Grid grid
    cpdef void set_params(self, dict params) except *

cdef class GridAgent(GridItem):
    cdef public long category
    cpdef rand_move_agent(self, int x_range, int y_range) except *

cdef class Spot(GridItem):
    cdef public long colormap

cdef class AgentIDManager:
    cdef long _width
    cdef long _height
    cdef vector[cpp_set[long]] _agents
    cdef long _max_id
    cdef cpp_set[long] _empty_spots
    cdef bint allow_multi
    cdef public cpp_set[long] all_categories 
    cdef void _add_agent(self, long agent_id, long category, long x, long y) except *
    cpdef add_agent(self, long agent_id, long category, long x, long y) except *
    cdef void _remove_agent(self, long agent_id, long category, long x, long y) except *
    cpdef remove_agent(self, long agent_id, long category, long x, long y) except *

    cdef long _convert_to_1d(self, long x, long y)
    
    cdef cpp_set[long]* _get_empty_spots(self)

    cpdef set get_empty_spots(self)
    cpdef long agent_id_and_category_to_number(self, long agent_id, long category) except *
    cdef (long, long) num_to_2d_coor(self, long num) except *
    cpdef (long, long) number_to_category_and_agent_id(self, long num) except *

    cpdef list agents_on_spot(self, long x, long y) except *
    cdef (long, long) find_empty_spot(self) except *

cdef class Grid:
    cdef long _width
    cdef long _height
    cdef bint _multi
    cdef bint _caching
    cdef bint _wrap
    cdef public dict _existed_agents
    cdef list _spots
    cdef dict _agent_ids
    cdef dict _neighbors_cache
    cdef list _roles_list
    cdef list _agent_containers
    cdef set _agent_categories
    cdef public object scenario
    cdef public object _spot_cls
    
    cdef AgentIDManager _agent_id_mgr
    cpdef _add_agent_container(self, object category , str initial_placement) except *
    cpdef AgentList get_agent_container(self, category_id) except *
    cpdef Spot get_spot(self, long x, long y)
    cdef long _convert_to_1d(self, long x, long y)
    cdef bint _in_bounds(self, long x, long y)
    cpdef list _get_agent_id_set_list(self, object category_name)
    cdef (long, long) _bound_check(self, long x, long y) except *
    cdef (long, long) _coords_wrap(self, long x ,long y)
    cpdef (long, long) coords_wrap(self, long x, long y)

    cdef list _neighbors(self, long x, long y, long radius, bint moore, bint except_self)
    cdef long _get_neighbors_array_length(self, long radius, bint moore, bint except_self)
    cdef long _get_neighbors_key_hash(self, long x, long y, long radius, bint moore, bint except_self)
    cpdef list _get_neighbor_positions(self, long x, long y, long radius=*, bint moore=*, bint except_self=*)
    

    cpdef long height(self)
    cpdef long width(self)
    
    cpdef void add_agent(self, GridAgent agent) except *
    cpdef void remove_agent(self, GridAgent agent) except *
    cpdef void move_agent(self, GridAgent agent, long target_x, long target_y) except *
    cpdef list get_spot_agents(self, Spot spot) except *
    cdef void _add_agent(self, long agent_id, long category, long x, long y) except *
    cdef void _remove_agent(self, long agent_id, long category,long x, long y) except *
    cpdef list get_neighbors(self, GridAgent agent, long radius=*, bint moore=*, bint except_self=*) except *
    
    cpdef (long, long) rand_move_agent(self, GridAgent agent, long category, long range_x, long range_y) except *
    cpdef tuple get_colormap(self)
    cpdef (long, long) find_empty_spot(self) except *
    cpdef list get_empty_spots(self) except *


from Melodie import Model, Scenario
import pandas as pd
import typing
from typing import (
    Callable,
    List,
    Dict,
    Union,
    Set,
    Optional,
    TypeVar,
    Generic,
    Iterable,
)

from collections.abc import Sequence

AgentGeneric = TypeVar("AgentGeneric")
ContainerGeneric = TypeVar("ContainerGeneric")

class SeqIter(Generic[AgentGeneric]):
    def __init__(self, seq: Iterable[AgentGeneric]):
        self._seq = seq
        self._i = 0
    def __next__(self) -> AgentGeneric: ...

class BaseAgentContainer(Iterable[Generic[AgentGeneric]]):
    _agent_ids = set()

    def __init__(self):
        self._id_offset = -1
        self.scenario: Union[Scenario, None] = None
        self.agents: Union[List[AgentGeneric], Set[AgentGeneric], None] = None
    def __iter__(self) -> Generic[AgentGeneric]: ...
    def new_id(self) -> int: ...
    def all_agent_ids(self) -> List[int]: ...
    def to_list(self, column_names: List[str] = None) -> List[Dict]: ...
    def type_check(self, param_names: List[str], agent_params_df: pd.DataFrame): ...
    def set_properties(self, props_df: pd.DataFrame) -> None: ...
    def get_agent(self, agent_id: int) -> Optional[AgentGeneric]: ...

class AgentDict(BaseAgentContainer):
    pass

class AgentList(BaseAgentContainer, Sequence, typing.Sequence[AgentGeneric]):
    def __init__(self, agent_class: typing.Type[AgentGeneric], model: "Model") -> None:
        super(AgentList, self).__init__()
        self._iter_index = 0
        self.scenario = model.scenario
        self.agent_class: typing.Type[AgentGeneric] = agent_class
        self.initial_agent_num: int = -1
        self.model = model
        self.agents: List[AgentGeneric] = self.init_agents()
    def __repr__(self):
        return f"<AgentList {self.agents}>"
    def __len__(self):
        return len(self.agents)
    def __getitem__(self, item) -> AgentGeneric:
        return self.agents.__getitem__(item)
    def __iter__(self) -> SeqIter: ...
    def setup(self) -> None: ...
    def init_agents(self) -> List[AgentGeneric]: ...
    def setup_agents(self, agents_num: int, params_df: pd.DataFrame = None): ...
    def random_sample(self, sample_num: int) -> List["AgentGeneric"]: ...
    def remove(self, agent: "AgentGeneric") -> None: ...
    def add(self, agent: "AgentGeneric" = None, params: Dict = None) -> None: ...
    def to_list(self, column_names: List[str] = None) -> List[Dict]: ...
    def to_dataframe(self, column_names: List[str] = None) -> pd.DataFrame: ...
    def set_properties(self, props_df: pd.DataFrame): ...
    def filter(
        self, condition: Callable[["AgentGeneric"], bool]
    ) -> List["AgentGeneric"]: ...



# DateTimeProperty, DateProperty, and TimeProperty
# include code copied from couchdbkit
from __future__ import absolute_import

import inspect
import sys
import datetime
import time
import decimal
from .base_properties import (
    AbstractDateProperty,
    AssertTypeProperty,
    JsonContainerProperty,
    JsonProperty,
    DefaultProperty,
)
from .containers import JsonArray, JsonDict, JsonSet


if sys.version > '3':
    unicode = str
    long = int


class StringProperty(AssertTypeProperty):
    _type = (unicode, str)

    def selective_coerce(self, obj):
        if isinstance(obj, str):
            obj = unicode(obj)
        return obj


class BooleanProperty(AssertTypeProperty):
    _type = bool


class IntegerProperty(AssertTypeProperty):
    _type = (int, long)


class FloatProperty(AssertTypeProperty):
    _type = float

    def selective_coerce(self, obj):
        if isinstance(obj, (int, long)):
            obj = float(obj)
        return obj


class DecimalProperty(JsonProperty):

    def wrap(self, obj):
        return decimal.Decimal(obj)

    def unwrap(self, obj):
        if isinstance(obj, (int, long)):
            obj = decimal.Decimal(obj)
        elif isinstance(obj, float):
            # python 2.6 doesn't allow a float to Decimal
            obj = decimal.Decimal(unicode(obj))
        assert isinstance(obj, decimal.Decimal)
        return obj, unicode(obj)


class DateProperty(AbstractDateProperty):

    _type = datetime.date

    def _wrap(self, value):
        fmt = '%Y-%m-%d'
        try:
            return datetime.date(*time.strptime(value, fmt)[:3])
        except ValueError as e:
            raise ValueError('Invalid ISO date {0!r} [{1}]'.format(value, e))

    def _unwrap(self, value):
        return value, value.isoformat()


class DateTimeProperty(AbstractDateProperty):

    _type = datetime.datetime

    def _wrap(self, value):
        if not self.exact:
            value = value.split('.', 1)[0]  # strip out microseconds
            value = value[0:19]  # remove timezone
            fmt = '%Y-%m-%dT%H:%M:%S'
        else:
            fmt = '%Y-%m-%dT%H:%M:%S.%fZ'
        try:
            return datetime.datetime.strptime(value, fmt)
        except ValueError as e:
            raise ValueError(
                'Invalid ISO date/time {0!r} [{1}]'.format(value, e))

    def _unwrap(self, value):
        if not self.exact:
            value = value.replace(microsecond=0)
            padding = ''
        else:
            padding = '' if value.microsecond else '.000000'
        return value, value.isoformat() + padding + 'Z'


class TimeProperty(AbstractDateProperty):

    _type = datetime.time

    def _wrap(self, value):
        if not self.exact:
            value = value.split('.', 1)[0]  # strip out microseconds
            fmt = '%H:%M:%S'
        else:
            fmt = '%H:%M:%S.%f'
        try:
            return datetime.time(*time.strptime(value, fmt)[3:6])
        except ValueError as e:
            raise ValueError('Invalid ISO time {0!r} [{1}]'.format(value, e))

    def _unwrap(self, value):
        if not self.exact:
            value = value.replace(microsecond=0)
        return value, value.isoformat()


class ObjectProperty(JsonProperty):

    _type = None
    default = lambda self: self.item_type()

    def __init__(self, item_type=None, **kwargs):
        self._item_type_deferred = item_type
        super(ObjectProperty, self).__init__(**kwargs)

    @property
    def item_type(self):
        from .base import JsonObjectBase
        if hasattr(self, '_item_type_deferred'):
            if inspect.isfunction(self._item_type_deferred):
                self._item_type = self._item_type_deferred()
            else:
                self._item_type = self._item_type_deferred
            del self._item_type_deferred
        if not issubclass(self._item_type, JsonObjectBase):
            raise ValueError("item_type {0!r} not a JsonObject subclass".format(
                self._item_type,
                self.type_config.properties,
            ))
        return self._item_type

    def wrap(self, obj, string_conversions=None):
        return self.item_type.wrap(obj)

    def unwrap(self, obj):
        assert isinstance(obj, self.item_type), \
            '{0} is not an instance of {1}'.format(obj, self.item_type)
        return obj, obj._obj


class ListProperty(JsonContainerProperty):

    _type = default = list
    container_class = JsonArray

    def _update(self, container, extension):
        container.extend(extension)


class DictProperty(JsonContainerProperty):

    _type = default = dict
    container_class = JsonDict

    def _update(self, container, extension):
        container.update(extension)


class SetProperty(JsonContainerProperty):

    _type = default = set
    container_class = JsonSet

    def _update(self, container, extension):
        container.update(extension)



import time

import pandas as pd


class MyData:
    def __init__(self, df: pd.DataFrame):
        self.records = df.to_dict("records")
        self.indices = {
            "id": {record["id"]: {row} for row, record in enumerate(self.records)}
        }

    def get_data(self, **kwargs):
        s: set = None
        for arg_name, arg_value in kwargs.items():
            indices = self.indices[arg_name][arg_value]
            if s is None:
                s = indices
            else:
                s = s.intersection(indices)
        records = []
        for row in s:
            records.append(self.records[row])
        if len(records) == 1:
            return records[0]
        return records


df = pd.DataFrame([{"a": 1, "b": 1, "id": i} for i in range(10000)])
df.set_index(["id"])

M = 100000


def df_getitem_speed():
    s = 0
    t0 = time.time()
    for i in range(M):
        a = df.loc[9999, "b"]
        # s += a
    print("df:", time.time() - t0)


def mycls_getitem_speed():
    s = 0
    my_data = MyData(df)
    t0 = time.time()
    for i in range(M):
        a = my_data.get_data(id=9999)
        s += a["b"]
    print("mycls", time.time() - t0)


df_getitem_speed()
mycls_getitem_speed()



import os
from typing import Any, Dict


class Config:
    """
    The configuration class of Melodie
    Config is needed by Simulator/Calibrator/Trainer and MelodieStudio for determining the project root,
    IO directories or other crucial configurations.
    """

    def __init__(
            self,
            project_name: str,
            project_root: str,
            input_folder: str,
            output_folder: str,
            visualizer_entry: str = "",
            **kwargs,
    ):
        self.project_name = project_name
        self.project_root = project_root
        self.output_folder = self.setup_folder_path(output_folder)
        self.input_folder = self.setup_folder_path(input_folder)
        self.temp_folder = ".melodie"

        self.studio_port = kwargs.get("studio_port", 8089)
        self.visualizer_port = kwargs.get("visualizer_port", 8765)
        self.parallel_port = kwargs.get("parallel_port", 12233)

        if not os.path.exists(visualizer_entry) and visualizer_entry != "":
            raise FileNotFoundError(
                f"Visualizer entry file {visualizer_entry} is defined, but not found. "
            )
        self.visualizer_entry = visualizer_entry
        self.visualizer_tmpdir = os.path.join(self.temp_folder, "visualizer")
        self.init_temp_folders()

        self.setup()

    def init_temp_folders(self):
        if not os.path.exists(self.temp_folder):
            os.makedirs(self.temp_folder)
        if not os.path.exists(self.visualizer_tmpdir):
            os.makedirs(self.visualizer_tmpdir)

    def setup_folder_path(self, folder_path):
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        return folder_path

    def setup(self):
        pass

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_dict(d: Dict[str, Any]):

        c = Config(
            d["project_name"],
            d["project_root"],
            d["input_folder"],
            d["output_folder"],
        )
        return c



# -*- coding:utf-8 -*-
# @Time: 2022/11/9 10:41
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: params.py.py
import math
from typing import Callable, Any, TYPE_CHECKING, Tuple, Union, List, Dict

from .vis_charts import JSONBase

if TYPE_CHECKING:
    from Melodie import Scenario

GetterType = 'Callable[[object], Any]'
SetterType = 'Callable[[object, Param], Any]'
GetterArgType = Union[GetterType, str]
SetterArgType = Union[SetterType, str]

BasicType = Union[str, int, float, bool]
ParamsType = 'Union[Param, IntParam, FloatParam, BoolParam, StringParam, ArrayParam]'


class UnInitialized:
    """
    Special class standing for uninitialized variable

    """
    pass


class Param(JSONBase):
    """
    Basic Parameter object

    """

    def validate_getter_or_setter(self, f):
        assert isinstance(f, str) or callable(f), 'Argument `getter` and `setter` should be callable or string'

    def __init__(self, name: str, getter: GetterArgType, setter: SetterArgType, readonly=False, label="",
                 description="No description about this parameter...",
                 component='auto'):
        self.name = name
        self.label = label
        self.type = ''
        self.component: str = component
        if isinstance(getter, str) and getter == '':
            getter = name
        if isinstance(setter, str) and setter == '':
            setter = name
        self.validate_getter_or_setter(getter)
        self.validate_getter_or_setter(setter)
        self._value: Any = UnInitialized()
        self.readonly = readonly
        self.description = description

        self._getter_from_obj: GetterType = getter if callable(getter) else \
            lambda obj: getattr(obj, getter)
        self._setter_to_obj: SetterType = setter if callable(setter) else \
            lambda obj, param: setattr(obj, setter, param)

    @property
    def value(self):
        """
        Get Value

        :return:
        """
        return self._value

    def children(self):
        return None

    @value.setter
    def value(self, new_val):
        new_val = self._converter(new_val)
        self._validator(new_val)
        assert isinstance(new_val, type(self._value)) or isinstance(self._value, UnInitialized), (
            'new value', type(new_val), 'old value', type(self._value), 'does not match')
        self._value = new_val

    def to_json(self):
        """
        Convert param structure to json

        :return:
        """
        d = super().to_json()
        return d

    def to_value_json(self):
        """
        Convert param value to json

        :return:
        """
        return {'value': self._value, 'type': self.type, 'name': self.name}

    def _validator(self, new_val):
        return

    def _converter(self, new_val):
        """
        Convert any valid value to integer.

        :param new_val:
        :return:
        """
        return new_val

    def extract_value(self, obj: "object"):
        """
        Get value from object with getter

        :param obj:
        :return:
        """
        self.value = self._getter_from_obj(obj)

    def modify_object(self, obj: 'object'):
        """
        Modify property inside the object with setter.

        :param obj:
        :return:
        """
        self._setter_to_obj(obj, self.value)

    def from_json(self, d: Dict):
        """
        Load this object from json-dumpable(parsable) dict object

        :param d:
        :return:
        """
        self.value = d['value']


class IntParam(Param):
    """
    Integer parameter.

    Notice:

    The parameter `value_range` indicates a **closed** interval.

    For example, if `value_range` is a tuple `(0, 100)`, both 0 and 100 are valid values.

    Be sure to check the right boundary in case of out-of-range error when you are using this value as indices.

    """
    _value: int

    def __init__(self, name: str, value_range: Tuple[int, int], getter: GetterArgType = "", setter: SetterArgType = "",
                 readonly=False, label="",
                 description="",
                 component='auto', ):
        super().__init__(name, getter, setter, readonly, label, description, component)

        self.min = value_range[0]
        self.max = value_range[1]
        self.type = 'int'

    def _validator(self, new_val: int):
        """
        Int value should be in range

        :param new_val:
        :return:
        """
        if not self.min <= new_val <= self.max:
            raise ValueError(f"Integer value {new_val} out of range {self.min}<= x <={self.max}")

    def _converter(self, new_val):
        return int(new_val)


class BoolParam(Param):
    """
    Boolean parameter

    """
    _value: bool

    def __init__(self, name: str, getter: GetterArgType = "", setter: SetterArgType = "",
                 readonly=False, label="",
                 description="",
                 component='auto', ):
        super().__init__(name, getter, setter, readonly, label, description, component)
        self.type = 'bool'

    def _converter(self, new_val):
        return bool(new_val)


class StringParam(Param):
    """
    String parameter
    
    """
    _value: bool

    def __init__(self, name: str, getter: GetterArgType = "", setter: SetterArgType = "",
                 readonly=False, label="",
                 description="",
                 component='auto', ):
        super().__init__(name, getter, setter, readonly, label, description, component)
        self.type = 'str'

    def _converter(self, new_val):
        return str(new_val)


class FloatParam(Param):
    """
    Parameter with float value.

    precision: An integer for float digit numbers, -1 by default, indicating no rounding.
    """
    _value: float

    def __init__(self, name: str, value_range: Tuple[float, float], step: float = -1,
                 getter: GetterArgType = "",
                 setter: SetterArgType = "",
                 readonly=False, label="",
                 description="",
                 component='auto', ):
        super().__init__(name, getter, setter, readonly, label, description, component)

        self.min = value_range[0]
        self.max = value_range[1]
        self.step = step
        self.type = 'float'

    def _validator(self, new_val: int):
        if not self.min - 10 ** -9 <= new_val <= self.max + 10 ** -9:
            raise ValueError(
                f"Float parameter named {self.name}, value {new_val} out of range {self.min}<= x <={self.max}")

    def _converter(self, new_val):
        return float(new_val)


class ArrayParam(Param):
    """
    Parameters could be set inside an array

    """
    _value: "List[Union[IntParam, FloatParam, ArrayParam]]"

    def __init__(self, name: str, value: "List[Union[IntParam, FloatParam, ArrayParam]]", getter: GetterArgType = '',
                 setter: SetterArgType = '', readonly=False, label="", description="", component='auto'):
        super().__init__(name, getter, setter, readonly, label, description,
                         component)
        self._value = value
        self.type = 'array'
        for val in self._value:
            assert isinstance(val, (IntParam, FloatParam, ArrayParam, BoolParam, StringParam))

    def children(self):
        """
        Get the parameters in the array.

        :return:
        """
        return self._value

    @property
    def value(self):
        """
        Getter of value, a list.

        :return:
        """
        return [param.value for param in self._value]

    @value.setter
    def value(self, new_val: List[BasicType]):
        """
        Setter of value

        :param new_val:
        :return:
        """
        assert len(new_val) == len(self._value)
        for i, val in enumerate(self._value):
            val.value = new_val[i]

    def to_json(self):
        """
        Convert to json

        :return:
        """
        items = []
        for param in self._value:
            items.append(param.to_json())
        d = super().to_json()
        d['children'] = items
        return d

    def to_value_json(self):
        """
        Convert to value json.

        :return:
        """
        return {"type": 'array', "name": self.name, "value": [param.to_value_json() for param in self._value]}

    def extract_value(self, obj: "object"):
        """
        Extract value from each child

        :param obj:
        :return:
        """
        for param in self._value:
            param.extract_value(obj)

    def from_json(self, d: List[Union[List, Dict]]):
        """
        Set value of each child from a json-dumpable list

        :param d:
        :return:
        """
        assert isinstance(d['value'], list)
        assert len(d['value']) == len(self._value)
        for i, param in enumerate(self._value):
            param.from_json(d['value'][i])

    def modify_object(self, obj: 'object'):
        """
        Modify each child with theirs setter

        :param obj:
        :return:
        """
        for param in self._value:
            param.modify_object(obj)


class ParamsManager:
    """
    Manager of parameters

    The parameter settings panel will be shown top-down on the left side of simulator page, and all parameter object
    added by `add_param` method will be put in a top-down layout.

    """

    def __init__(self):
        self._initialized = False
        self._unique_name_map: Dict[str, Param] = {}
        self.params: List[ParamsType] = []

    def add_param(self, param: ParamsType):
        """
        Add a parameter onto the panel.

        :param param:
        :return:
        """
        self.params.append(param)

    def to_json(self):
        """
        Convert the contained parameters to json

        :return:
        """
        return [param.to_json() for param in self.params]

    def modify_scenario(self, scenario: 'Scenario'):
        """
        Write parameter values to scenario object

        :param scenario:
        :return:
        """
        for param in self.params:
            param.modify_object(scenario)

    @staticmethod
    def write_obj_attrs_to_params_list(obj: object, params: List[ParamsType]):
        """
        Extract property value of object to parameters.

        :param obj:
        :param params:
        :return:
        """
        for param in params:
            param.extract_value(obj)

    @staticmethod
    def for_each_param(params: List[ParamsType], name_prefix: str, callback: Callable[[str, Param], None]):
        """
        Walk through the parameter tree to gather each parameter value into a plain list.

        :param params:
        :param name_prefix:
        :param callback:
        :return:
        """
        for param in params:
            assert isinstance(param, Param), param
            abs_name = name_prefix + "." + param.name
            if isinstance(param, ArrayParam):
                ParamsManager.for_each_param(param.children(), abs_name, callback)
            else:
                callback(abs_name, param)

    def to_form_model(self):
        """
        Convert data model to a json-dumpable list.

        :return:
        """
        return [param.to_json() for param in self.params]

    def to_value_json(self):
        """
        Convert value model to a json-dumpable list.

        :return:
        """
        return [param.to_value_json() for param in self.params]

    def from_json(self, l: List[Union[Dict, List]]):
        """
        Extract value of parameters from a json-dumpable list

        :param l:
        :return:
        """
        assert len(l) == len(self.params)
        for i, param in enumerate(self.params):
            param.from_json(l[i])



from typing import Set, List, Tuple, TYPE_CHECKING, Optional, Type

import numpy as np

from .. import AgentList, Scenario

if TYPE_CHECKING:
    from .basics import Agent


class AgentIDManager:
    pass


class GridItem(Agent):
    x: int
    y: int
    grid: Grid

    def __init__(self, agent_id: int, x: int = 0, y: int = 0):
        pass


class GridAgent(GridItem):
    category: int

    def __init__(self, agent_id: int, x: int = 0, y: int = 0, grid: Grid = None):
        super().__init__(agent_id, x, y)

    def setup(self):
        pass

    def rand_move_agent(self, x_range: int, y_range: int):
        pass

    def set_category(self):
        pass


class Spot(GridItem):
    role: int

    def __init__(self, spot_id: int, x: int = 0, y: int = 0):
        super(Spot, self).__init__(spot_id)

    def setup(self):
        pass

    def get_spot_agents(self) -> List[Tuple[int, int]]:
        pass


SpotGeneric = TypeVar("SpotGeneric")


class Grid(typing.Sequence[AgentGeneric]):
    scenario: "Scenario"
    wrap: bool

    def __init__(self, spot_cls: Type[Spot], scenario: Optional[Scenario] = None): ...

    def setup_params(
            self, width: int, height: int, wrap=True, caching=True, multi=True
    ): ...

    def init_grid(self): ...

    def setup(self) -> None: ...

    def set_spot_property(self, attr_name: str, array_2d): ...

    def setup_agent_locations(
            self, agents: "AgentList", initial_placement: str = "none"
    ): ...

    def width(self) -> int: ...

    def height(self) -> int: ...

    def get_spot(self, x, y) -> "SpotGeneric": ...

    def get_spot_agents(self, spot: Spot) -> List[Tuple[int, int]]: ...

    def coords_wrap(self, x, y) -> Tuple[int, int]: ...

    def get_neighbors(
            self, agent: GridAgent, radius: int = 1, moore=True, except_self=True
    ) -> List[Tuple[int, int]]: ...

    def _get_neighbor_positions(
            self, x: int, y: int, radius=1, moore=True, except_self=True
    ) -> List[Tuple[int, int]]: ...

    def add_agent(self, agent: GridAgent) -> None: ...

    def remove_agent(self, agent: GridAgent) -> None: ...

    def move_agent(self, agent: GridAgent, target_x, target_y) -> None: ...

    def to_2d_array(self, attr_name: str) -> np.ndarray: ...

    def get_colormap(self): ...

    def rand_move_agent(
            self, agent: GridAgent, category: int, x_range: int, y_range: int
    ) -> Tuple[int, int]: ...

    def find_empty_spot(self) -> Tuple[int, int]: ...

    def _get_neighborhood(
            self, x: int, y: int, radius: int = 1, moore=True, except_self=True
    ) -> List[Spot]: ...

    def get_empty_spots(self) -> Set[Tuple[int, int]]: ...

    def get_agent_neighborhood(
            self, agent: GridAgent, radius=1, moore=True, except_self=True
    ): ...

    def get_spot_neighborhood(
            self, spot: SpotGeneric, radius=1, moore=True, except_self=True
    ): ...

    @property
    def get_agent_categories(self) -> Set[int]:
        ...

    def get_agent_container(self, category_id) -> AgentList:
        ...



# -*- coding:utf-8 -*-
# @Time: 2022/11/23 21:29
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: __init__.py.py
from .table_files import *
from .web import DataServiceStatus, DataServiceState
from .database import DatabaseQueryRequest, DatabaseBasicRequest
from .typeutils import NUMERICAL_TYPE



# cython:language_level=3
# -*- coding:utf-8 -*-


from cpython.ref cimport PyObject  # somewhere at the top
from cpython cimport PyObject_GetAttr, PyObject_GetAttrString, \
    PyObject_GetItem, PyList_GetItem, PyList_Size, PyObject_SetAttr
cimport cython

cimport numpy as np
import numpy as np

ctypedef np.int64_t DTYPE_t
ctypedef np.float64_t DTYPE_FLOAT

ctypedef fused DTYPE_FUSED:
    np.int64_t
    np.float64_t


cpdef vectorize2d_int(object lst, object attr):
    cdef int x
    cdef int y
    cdef object inner_list
    cdef object obj2

    cdef Py_ssize_t dim1_size
    cdef Py_ssize_t dim2_size

    cdef np.ndarray[DTYPE_t, ndim=2] ret
    cdef DTYPE_t[:,:] result_view

    cdef object attr_val
    cdef PyObject* py_obj_ptr

    py_obj_ptr = <PyObject *>lst
    py_obj_ptr.ob_refcnt

    dim1_size = PyList_Size(lst)
    inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (0))
    dim2_size = PyList_Size(inner_list)
    ret = np.zeros([dim1_size, dim2_size], dtype=np.int64)
    result_view = ret
    for i in range(dim1_size):
        inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (i))
        for j in range(dim2_size):
            obj2 = <object> PyList_GetItem(inner_list, <Py_ssize_t> (j))
            attr_val = <object> PyObject_GetAttr(obj2,attr)
            result_view[i][j] = <DTYPE_t> attr_val
    return ret

cpdef vectorize2d_float(object lst, object attr):
    cdef int x, y
    cdef object inner_list, obj2

    cdef Py_ssize_t dim1_size, dim2_size

    cdef np.ndarray[DTYPE_FLOAT, ndim=2] ret
    cdef DTYPE_FLOAT[:,:] result_view

    cdef object attr_val
    cdef PyObject* py_obj_ptr

    py_obj_ptr = <PyObject *>lst
    py_obj_ptr.ob_refcnt

    dim1_size = PyList_Size(lst)
    inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (0))
    dim2_size = PyList_Size(inner_list)
    ret = np.zeros([dim1_size, dim2_size], dtype=np.float64)
    result_view = ret
    for i in range(dim1_size):
        inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (i))
        for j in range(dim2_size):
            obj2 = <object> PyList_GetItem(inner_list, <Py_ssize_t> (j))
            attr_val = <object> PyObject_GetAttr(obj2,attr)
            result_view[i][j] = <DTYPE_FLOAT> attr_val
    return ret


cpdef apply2d_int(object lst, object attr, np.ndarray new_attr):
    cdef int x
    cdef int y
    cdef object inner_list
    cdef object obj2

    cdef Py_ssize_t dim1_size
    cdef Py_ssize_t dim2_size

    cdef DTYPE_t[:,:] new_attr_view

    cdef object attr_val
    cdef PyObject* py_obj_ptr

    new_attr_view = new_attr
    py_obj_ptr = <PyObject *>lst
    py_obj_ptr.ob_refcnt

    dim1_size = PyList_Size(lst)
    inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (0))
    dim2_size = PyList_Size(inner_list)


    for i in range(dim1_size):
        inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (i))
        for j in range(dim2_size):
            obj2 = <object> PyList_GetItem(inner_list, <Py_ssize_t> (j))
            PyObject_SetAttr(obj2, attr, <object> new_attr_view[i][j])

cpdef apply2d_float(object lst, object attr, np.ndarray new_attr):
    cdef int x
    cdef int y
    cdef object inner_list
    cdef object obj2

    cdef Py_ssize_t dim1_size
    cdef Py_ssize_t dim2_size

    cdef DTYPE_FLOAT[:,:] new_attr_view

    cdef object attr_val
    cdef PyObject* py_obj_ptr

    new_attr_view = new_attr
    py_obj_ptr = <PyObject *>lst
    py_obj_ptr.ob_refcnt

    dim1_size = PyList_Size(lst)
    inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (0))
    dim2_size = PyList_Size(inner_list)


    for i in range(dim1_size):
        inner_list = <object> PyList_GetItem(lst, <Py_ssize_t> (i))
        for j in range(dim2_size):
            obj2 = <object> PyList_GetItem(inner_list, <Py_ssize_t> (j))
            PyObject_SetAttr(obj2,attr,<object> new_attr_view[i][j])



# cython:language_level=3
# distutils: language = c++
# cython: profile=False
# -*- coding:utf-8 -*-

cdef extern from "Python.h":
    const char* PyUnicode_AsUTF8(object unicode)

import functools
from typing import Type, Set, Dict, List, Tuple
from Melodie.boost.vectorize import vectorize_2d
from cpython.ref cimport PyObject
from cpython cimport PyObject_GetAttr, PyObject_GetAttrString, \
    PyObject_GetItem, PyList_GetItem, PyList_Size, PyObject_SetAttr
cimport cython
from libc.math cimport pow
from libc.stdlib cimport rand, RAND_MAX
from libcpp.vector cimport vector
from libcpp.unordered_set  cimport unordered_set as cpp_set
from libcpp.unordered_map cimport unordered_map as cpp_map
from libcpp.string cimport string as cpp_string
from libcpp.pair cimport pair as cpp_pair
from cython.operator cimport dereference as deref, preincrement as inc

from .basics cimport Element, Agent
from .agent_list cimport AgentList
cimport numpy as np
import random
import numpy as np
import time
ctypedef np.int64_t DTYPE_t
ctypedef np.float64_t DTYPE_FLOAT

ctypedef fused DTYPE_FUSED:
    np.int64_t
    np.float64_t

cdef class GridItem(Agent):
    def __init__(self, agent_id:int, grid, x:int=0, y:int=0 ):
        super().__init__(agent_id)
        self.grid = grid
        self.x = x
        self.y = y

    cpdef void set_params(self, dict params) except *:
        """

        :param params:
        :return:
        """
        for paramName, paramValue in params.items():
            assert hasattr(self, paramName), f"property named {paramName} not in {self.__class__.__name__}"
            setattr(self, paramName, paramValue)

    def __repr__(self):
        return f"<{self.__class__.__name__} 'x': {self.x}, 'y': {self.y}>"

cdef class GridAgent(GridItem):
    def __init__(self, agent_id: int, x: int = 0, y: int = 0, grid = None):
        super().__init__(agent_id, grid, x, y)
        self.category = -1
        self.set_category()
        assert self.category >= 0, "Category should be larger or "
    
    def set_category(self) -> int:
        """
        Set the category of GridAgent.

        As there may be more than one types of agent wandering around the grid, `category` is used to tell the type of
        `GridAgent`. So be sure to inherit this method in custom GridAgent implementation.

        :return: int
        """
        raise NotImplementedError("Category should be set for GridAgent")

    cpdef rand_move_agent(self, int x_range, int y_range) except *:
        """
        Randomly move to a new position within x and y range.

        :return: None
        """
        if self.grid is None:
            raise ValueError("Grid Agent has not been registered onto the grid!")
        self.x, self.y = self.grid.rand_move_agent(self, self.category, x_range, y_range)

cdef class Spot(GridItem):
    def __init__(self, spot_id: int, grid: Grid, x: int = 0, y: int = 0):
        super().__init__(spot_id, grid, x, y)
        self.grid = grid
        self.colormap = 0

    def get_spot_agents(self):
        """
        Get all agents on the spot.

        :return: a list of grid agent.
        """
        return self.grid.get_spot_agents(self)

    def __repr__(self):
        return f"<{self.__class__.__name__} 'x': {self.x}, 'y': {self.y}, 'colormap': {self.colormap}, 'payload' : {self.__dict__}>"

    def get_style(self):
        return {
            "backgroundColor": "#ffffff"
        }


cdef class AgentIDManager:
    def __init__(self, long width, long height, bint allow_multi=False):
        self._agents = [set() for i in range(width*height)]
        self._width = width
        self._height = height
        self._max_id = 1000000
        self.allow_multi = allow_multi
        self._empty_spots = set()
        self.all_categories = set()
        for x in range(self._width):
            for y in range(self._height):
                self._empty_spots.insert(self._convert_to_1d(x, y))

    cpdef add_agent(self, long agent_id, long category, long x, long y) except *:
        self._add_agent(agent_id, category, x, y)

    cdef void _add_agent(self, long agent_id, long category, long x, long y) except *:
        cdef long pos_1d = self._convert_to_1d(x, y)
        cdef cpp_set[long]* agents_on_spot = &self._agents[pos_1d]
        cdef long agent_num_repr = self.agent_id_and_category_to_number(agent_id, category)
        if (not self.allow_multi) and (agents_on_spot.count(pos_1d)!=0):  # If not allow multi, the spot should be empty before adding an agent.
            raise ValueError(f"Multiple agents on one spot is not allowed.")
        if agents_on_spot.count(agent_num_repr)!=0:
            raise ValueError(f"Agent id {agent_id}, category {category} already on spot <{x}, {y}>")
        
        self._empty_spots.erase(pos_1d)
        agents_on_spot.insert(agent_num_repr)
        self.all_categories.insert(category)
    
    cpdef remove_agent(self, long agent_id, long category, long x, long y) except *:
        self._remove_agent(agent_id, category, x, y)

    cdef void _remove_agent(self, long agent_id, long category, long x, long y) except *:
        cdef long pos_1d = self._convert_to_1d(x, y)
        cdef cpp_set[long]* agents_on_spot = &self._agents[pos_1d]
        cdef long agent_num_repr = self.agent_id_and_category_to_number(agent_id, category)
        if agents_on_spot.count(agent_num_repr)==0:
            raise ValueError(f"Agent id {agent_id}, category {category} does not exist on spot <{x}, {y}>")
        agents_on_spot.erase(agent_num_repr)
        if agents_on_spot.size()==0:
            self._empty_spots.insert(pos_1d)

    cdef long _convert_to_1d(self, long x, long y):
        return x * self._height + y

    cdef cpp_set[long]* _get_empty_spots(self):
        return &self._empty_spots

    cpdef set get_empty_spots(self):
        return self._empty_spots

    cpdef long agent_id_and_category_to_number(self, long agent_id, long category) except *:
        if not 0<=agent_id<=self._max_id:
            raise ValueError(f"Agent id {agent_id} out of range!")
        return category*self._max_id + agent_id

    @cython.cdivision(True)
    cpdef (long, long) number_to_category_and_agent_id(self, long num) except *:
        """
        return: (agent_id, category_id)
        """
        return num/self._max_id, num%self._max_id

    @cython.cdivision(True)
    cdef (long, long) num_to_2d_coor(self, long num) except *:
        return num/self._height, num%self._width

    cpdef list agents_on_spot(self, long x, long y) except *:
        cdef list l = []
        cdef long spot = 0
        for spot in self._agents[self._convert_to_1d(x, y)]:
            l.append(self.number_to_category_and_agent_id(spot))
        return l

    cdef (long, long) find_empty_spot(self) except *:
        cdef long rand_value = random.randint(0, self._empty_spots.size()-1)
        cdef long i = 0
        for item in self._empty_spots:
            if i==rand_value:
                return self.num_to_2d_coor(item)
            i+=1


cdef class Grid:
    """
    Grid is a widely-used discrete space for ABM. It contains many ``Spot``s,
    and agents move and interact inside spots.
    """

    def __init__(self, spot_cls: Type[Spot], scenario=None):
        """

        :param spot_cls: The class of Spot
        :param width: The width of Grid
        :param height: The height of Grid
        :param wrap: If true, the coordinate overflow will be mapped to another end.
        :param caching: If true, the neighbors and bound check results will be cached to avoid re-computing.

        """
        self.scenario = scenario
        self._spot_cls = spot_cls
        self._width = -1
        self._height = -1
        self._wrap = True
        self._multi = False
        self._caching = True

    def init_grid(self):
        """
        Initialize the grid.

        :return: None
        """
        self._spots = [[self._spot_cls(self._convert_to_1d(x, y), self, x, y) for x in range(self._width)] for y in range(self._height)]
        for x in range(self._width):
            for y in range(self._height):
                self._spots[y][x].setup()

        self._neighbors_cache = {}
        self._roles_list = [[0 for j in range(4)] for i in range(self._width*self._height)]
        self._agent_id_mgr = AgentIDManager(self._width, self._height, allow_multi=self._multi)
        self._agent_containers = [None for i in range(100)]
        self._agent_categories = set()
        

    def setup_params(self, width: int, height: int, wrap=True, caching=True, multi=True):
        """
        Setup the parameters of grid.

        :param width: int
        :param height: int
        :param wrap: bool, True by default.
        If True, GridAgent will re-enter the grid on the other side if it moves out of the grid on one side.
        :param caching: bool, True by default. If true, the grid caches the neighbor of each spot.
        :param multi: bool, True by default. If true, more than one agent could stand on one spot. If false, error will
        be raised when attempting to place multiple agents on one spot.
        :return: None
        """
        self._width = width
        self._height = height
        self._wrap = wrap
        self._caching = caching
        self._multi = multi
        self.init_grid()

    def setup(self):
        """
        Be sure to inherit this function.

        :return: None
        """
        pass    
    
    def _setup(self):
        self.setup()

    @property
    def get_agent_categories(self):
        return self._agent_categories

    def setup_agent_locations(self, category ,initial_placement = "direct")->None:
        """
        Add an agent category.
        
        For example, if there are two classes of agents: `Wolf(GridAgent)` and `Sheep(GridAgent)`, 
        and there are 100 agents with id 0~99 for each class. It is obvious in such a circumstance that 
        we cannot identify an agent only with agent *id*.So it is essential to use *category_name* to distinguish two types of agents. 

        :param category_id: The id of new category.
        :param category: An AgentList object
        :param initial_placement: A str object stand for initial placement.
        :return: None
        """
        initial_placement = initial_placement.lower()
        self._add_agent_container(category, initial_placement)

    cpdef _add_agent_container(self, object category, str initial_placement) except *:
        cdef GridAgent agent
        cdef tuple pos
        assert category is not None, f"Agent Container was None"
        agent = category[0]
        category_id = agent.category
        assert 0<=category_id<100, f"Category ID {category_id} should be a int between [0, 100)"
        assert self._agent_containers[category_id] is None, f"Category ID {category_id} already existed!"
        self._agent_containers[category_id] = category
        self._agent_categories.add(category_id)
        assert initial_placement in {"random_single", "direct"}, f"Invalid initial placement '{initial_placement}' "
        if initial_placement == "random_single":
            for agent in category:
                pos = self.find_empty_spot()
                agent.x = pos[0]
                agent.y = pos[1]
                self.add_agent(agent)
        elif initial_placement == "direct":
            for agent in category:
                self.add_agent(agent)

    cpdef Spot get_spot(self, long x, long y):
        """
        Get a ``Spot`` at position ``(x, y)``

        :param x:
        :param y:
        :return: The ``Spot`` at position (x, y)
        """
        cdef list row
        x, y = self._bound_check(x, y)
        
        row = <list>PyList_GetItem(self._spots, <Py_ssize_t> (y))
        return <Spot>PyList_GetItem(row, <Py_ssize_t> (x))

    cpdef list get_spot_agents(self, Spot spot) except *:
        """
        Get all agent of a specific category from the spot at ``(x, y)``

        :param category: category name of agent.
        :param x: 
        :param y:
        :return: A set of int, the agent ids.
        """
        return self._agent_id_mgr.agents_on_spot(spot.x, spot.y)


    cpdef AgentList get_agent_container(self, category_id) except *:
        cdef AgentList ret = self._agent_containers[category_id]
        assert ret is not None, f"Agent List for category id {category_id} is not registered!"
        return ret

    cdef long _convert_to_1d(self, long x, long y):
        return x * self._height + y

    cdef bint _in_bounds(self, long x, long y):
        return (0 <= x < self._width) and (0 <= y <= self._height)

    cpdef list _get_agent_id_set_list(self, object category_name):
        """
        Get category of agents_series_data
        """
        cdef list category
        category = self._agent_ids.get(category_name)
        if category is None:
            raise ValueError(f"Category {category_name} is not registered!")
        return category

    @cython.initializedcheck(False)
    cdef (long, long) _bound_check(self, long x, long y) except *:
        if self._wrap:
            return self._coords_wrap(x, y)
        if not (0 <= x < self._width):
            raise IndexError("grid index x was out of range")
        elif not (0 <= y <= self._height):
            raise IndexError("grid index y was out of range")
        else:
            return x, y

    @cython.cdivision(True)
    cdef (long, long) _coords_wrap(self, long x ,long y):
        """
        Wrap the coordination
        :param x:
        :param y:
        :return:
        """
        cdef long rem_x, rem_y
        rem_x = x % self._width
        rem_y = y % self._height 
        if rem_x < 0:
            rem_x += self._width
        if rem_y < 0:
            rem_y += self._height
        return rem_x, rem_y

    cpdef (long, long) coords_wrap(self, long x, long y):
        return self._coords_wrap(x, y)

    cdef long _get_neighbors_array_length(self, long radius, bint moore, bint except_self):
        cdef long length 
        length = 0
        if moore:
            length = (radius * 2 + 1) ** 2
        else:
            length = 2 * radius * (radius + 1) + 1
        if except_self:
            length -= 1
        return length
        
    cdef long _get_neighbors_key_hash(self, long x, long y, long radius, bint moore, bint except_self):
        """
        The key hash is:
        """
        cdef long ret
        ret =  x*2**22 
        ret += y*2**12
        ret += radius*2**3
        ret += (<long>moore)*2 + <long>except_self
        return ret

    cdef list _neighbors(self, long x, long y, long radius, bint moore, bint except_self):
        cdef long dx, dy, length, coor_x, coor_y, counter
        cdef list neighbors
        
        x, y = self._bound_check(x, y)
        length = self._get_neighbors_array_length(radius, moore, except_self)
        neighbors = [None]*length
        counter = 0
        for dx in range(-radius, radius + 1):
            for dy in range(-radius, radius + 1):
                if not moore and abs(dx) + abs(dy) > radius:
                    continue
                if not self._wrap and not self._in_bounds(x + dx, y + dy):
                    continue
                if dx == 0 and dy == 0 and except_self:
                    continue
                neighbors[counter] = self._bound_check(x + dx, y + dy)
                counter += 1
        return neighbors

    cpdef list _get_neighbor_positions(self, long x, long y, long radius = 1, bint moore=True, bint except_self=True):
        cdef list neighbors 
        cdef tuple key
        key = (x, y, radius, moore, except_self)
        if self._neighbors_cache.get(key) is not None:
            return self._neighbors_cache[key]
        else:
            neighbors = self._neighbors(x, y, radius, moore, except_self)
            self._neighbors_cache[key] = neighbors
            return neighbors

    cpdef list get_neighbors(self, GridAgent agent, long radius=1, bint moore=True, bint except_self=True) except *:
        """
        Get the neighbors of one spot at (x, y).

        :param x:
        :param y:
        :param radius:
        :param moore:
        :param except_self:
        :return:  A list of the tuple: (`Agent category`, `Agent id`).
        """
        neighbor_ids = []
        neighbor_positions = self._get_neighbor_positions(agent.x, agent.y, radius, moore, except_self)
        for neighbor_pos in neighbor_positions:
            x, y = neighbor_pos
            agent_ids = self.get_spot_agents(
                self.get_spot(x, y)
            )  
            neighbor_ids.extend(agent_ids)
        return neighbor_ids
    
    def _get_neighborhood(self, x, y, radius = 1, moore=True, except_self=True):
        """
        Get all spots around (x, y)

        """
        neighbor_positions = self._get_neighbor_positions(x, y, radius, moore, except_self)
        spots = []
        for pos in neighbor_positions:
            x, y = pos
            spots.append(self.get_spot(x, y))
        return spots

    def get_agent_neighborhood(self, agent, radius = 1, moore=True, except_self=True):
        return self._get_neighborhood(agent.x, agent.y, radius, moore, except_self)

    def get_spot_neighborhood(self, spot, radius = 1, moore=True, except_self=True):
        return self._get_neighborhood(spot.x, spot.y, radius, moore, except_self)

    cdef void _add_agent(self, long agent_id, long category, long x, long y) except *:
        """
        Add agent onto the grid

        :param agent_id:
        :param category:
        :param x:
        :param y:
        :return:
        """
        x, y = self._bound_check(x, y)
        self._agent_id_mgr._add_agent(agent_id, category, x, y)

    cdef void _remove_agent(self, long agent_id, long category,long x, long y) except *:
        x, y = self._bound_check(x, y)
        self._agent_id_mgr._remove_agent(agent_id, category, x, y)

    cpdef void add_agent(self, GridAgent agent) except *:
        """
        Add an agent to the grid

        :param agent: An GridAgent object.
        :param category: A string, the name of category. The category should be registered. 
        :return:
        """
        agent.grid = self
        self._add_agent(agent.id, agent.category, agent.x, agent.y)

    cpdef void remove_agent(self, GridAgent agent) except *:
        """
        Remove an agent from the grid

        :param agent: An GridAgent object.
        :param category: A string, the name of category. The category should be registered. 
        :return: None
        """
        agent.grid = None
        self._remove_agent(agent.id, agent.category, agent.x, agent.y)

    cpdef void move_agent(self, GridAgent agent, long target_x, long target_y) except *:
        """
        Move agent to target position.

        :param agent: An GridAgent object.
        :param category: A string, the name of category. The category should be registered. 
        :param target_x: The target x coordinate
        :param target_y: The target y coordinate
        :return: None
        """
        
        self._remove_agent(agent.id, agent.category, agent.x, agent.y)
        self._add_agent(agent.id, agent.category, target_x, target_y)
        agent.x = target_x
        agent.y = target_y

    @cython.cdivision(True)
    cpdef (long, long) rand_move_agent(self, GridAgent agent, long category, long range_x, long range_y) except *:
        """
        Randomly move an agent with maximum movement `range_x` in x axis and `range_y` in y axis.
        
        :param agent: Must be `Melodie.GridAgent`, not `Agent`. That is because `GridAgent` has predefined properties required in `Grid`. 
        :param range_x: The activity range of agent on the x axis. 
        :param range_y: The activity range of agent on the y axis.
        
        For example, if the agent is at `(0, 0)`, `range_x=1` and `range_y=0`, the result can be
        `(-1, 0), (0, 0) or (1, 0)`. The probability of these three outcomes are equal.

        :return: (int, int), the new position
        """
        cdef long source_x, source_y, dx, dy, target_x, target_y
        source_x = agent.x
        source_y = agent.y
        self._remove_agent(agent.id, category, source_x, source_y)
        dx = <long>((rand()/(RAND_MAX*1.0))*(2*range_x+1)) - range_x
        dy = <long>((rand()/(RAND_MAX*1.0))*(2*range_y+1)) - range_y
        target_x = source_x+dx
        target_y = source_y+dy
        self._add_agent(agent.id, category, target_x, target_y)
        return self._coords_wrap(target_x, target_y)

    def to_2d_array(self, attr_name: str) -> np.ndarray:
        """
        Collect property of each spot and write the property value into an 2d np.array.

        Notice:

        - The property to collect should be float/int/bool, not other types such as str.
        - If you would like to get an element from the returned array, please write like this:
        
        .. code-block:: python
            :linenos:

            arr = self.to_2d_array('some_attr')
            y = 10
            x = 5
            spot_at_x_5_y_10 = arr[y][x] # CORRECT. Get the some_attr value of spot at `x = 5, y = 10`
            spot_at_x_5_y_10 = arr[x][y] # INCORRECT. You will get the value of spot at `x = 10, y = 5`
        

        :param attr_name: the property name to collect for this model.
        :return:
        """
        return vectorize_2d(self._spots, attr_name)

    def set_spot_property(self, attr_name: str, array_2d):
        """
        Set property from an 2d-numpy-array to each spot.

        """
        assert len(array_2d.shape) == 2, f"The spot property array should be 2-dimensional, but got shape: {array_2d.shape}"
        assert len(array_2d) == self._height, f"The rows of spot property matrix is {len(array_2d)} while the height of grid is {self._height}."
        assert len(array_2d[0]) == self._width, f"The columns of spot property matrix is {len(array_2d[0])} while the width of grid is {self._width}."
        for y, row in enumerate(array_2d):
            for x, value in enumerate(row):
                spot = self.get_spot(x, y)
                setattr(spot, attr_name, value)

    cpdef tuple get_colormap(self):
        """
        Get the role of each spot.

        :return: A tuple. The first item is a nested list for spot roles, and the second item is a dict for agent roles.
        """
        cdef Spot spot
        cdef long pos_1d 
        cdef long x, y, i, j
        cdef long category_id, agent_category
        cdef list category
        cdef dict agents_series_data
        cdef set agent_id_set
        cdef list role_pos_list

        agents_series_data = {}
        for category_id in self._agent_id_mgr.all_categories:
            agents_series_data[category_id] = []
        for x in range(self._width):
            for y in range(self._height):
                spot = self.get_spot(x, y)
                pos_1d = self._convert_to_1d(x, y)
                role_pos_list = self._roles_list[pos_1d]
                role_pos_list[0] = x
                role_pos_list[1] = y
                role_pos_list[2] = 0
                role_pos_list[3] = spot.colormap
                for agent_category, agent_id in self._agent_id_mgr.agents_on_spot(x, y):
                    series_data_one_category = agents_series_data[agent_category]
                    series_data_one_category.append({
                        'value': [x, y],
                        'id': agent_id,
                        'category': agent_category,
                        })

        return self._roles_list, agents_series_data
    
    cpdef long height(self):
        """
        Get the height of grid 

        :return: height, an ``int``
        """
        return self._height
    
    cpdef long width(self):
        """
        Get the width of grid

        :return: width, an ``int``
        """
        return self._width

    cpdef (long, long) find_empty_spot(self) except *:
        """
        Get a coordinate (x, y) of a spot without any agent on it.

        :return: A tuple, (x, y)
        """
        return self._agent_id_mgr.find_empty_spot()

    cpdef list get_empty_spots(self) except *:
        """
        Get all empty spots from grid.

        :return: a list of empty spot coordinates.
        """
        cdef list positions = []
        for spot_pos_1d in self._agent_id_mgr.get_empty_spots():
            positions.append(self._agent_id_mgr.num_to_2d_coor(spot_pos_1d))
        return positions


import logging
from typing import Dict, Set, Union, List, Tuple, Type

import networkx as nx
import numpy as np

from .boost.basics import Agent
from .boost.agent_list import AgentList

logger = logging.getLogger(__name__)


class NetworkAgent(Agent):
    id: int
    category: int
    network: "Network"

    def _set_network(self, network: "Network"):
        self.network = network

    def set_category(self):
        """
        Set the category of network agent.

        :return:
        """
        raise NotImplementedError(
            "Method `set_category` should be implemented for custom NetworkAgent."
        )


class Edge:
    def __init__(
            self,
            category_1: int,
            agent_1_id: int,
            category_2: int,
            agent_2_id: int,
            edge_properties: Dict[str, Union[int, str, float, bool]],
    ):
        """
        
        :param category_1: Source agent category
        :param agent_1_id: Source agent id
        :param category_2: Target agent category
        :param agent_2_id: Target agent id
        :param edge_properties: a ``dict`` storing properties.
        :return: None
        """
        self.category_1 = category_1
        self.agent_1_id = agent_1_id
        self.category_2 = category_2
        self.agent_2_id = agent_2_id
        self.properties: Dict[str, Union[int, str, float, bool]] = edge_properties
        self.setup()
        self.post_setup()

    def setup(self):
        """
        Setup method. Be sure to inherit it for custom ``Edge``.

        :return: None
        """
        pass

    def post_setup(self):
        """
        This method is executed after `setup()`

        :return: None
        """
        for prop_name, prop_value in self.properties.items():
            setattr(self, prop_name, prop_value)

    def __repr__(self):
        return f"<{self.__class__.__name__} {(self.category_1, self.agent_1_id)} --> {(self.category_2, self.agent_2_id)}>"


NodeType = Tuple[int, int]


class Network:
    def __init__(self, edge_cls: Type[Edge] = None, directed=False):
        self.simple = True
        self.directed = directed
        self.nodes: Set[NodeType] = set()
        self.edges: Dict[NodeType, Dict[NodeType, Edge]] = {}
        self.edge_cls: Type[Edge] = edge_cls if edge_cls is not None else Edge
        self.agent_categories: Dict[int, AgentList] = {}

        self.layout = {}
        self._layout_creator = lambda G: nx.spring_layout(G)

        self.setup()

    def _setup(self):
        self.setup()

    def setup(self):
        """
        Setup function of network. Be sure to inherit this method for custom implementation.

        :return: None
        """
        pass

    def add_edge(self, source_id: NodeType, target_id: NodeType, edge: Edge):
        """
        Add an edge onto the network.

        :param source_id: A tuple ``<agent_category, agent_id>`` stands for source agent.
        :param target_id: A tuple ``<agent_category, agent_id>`` stands for target agent.
        :param edge: An ``Edge`` object
        :return: None
        """
        if source_id not in self.edges:
            self.edges[source_id] = {}
        self.edges[source_id][target_id] = edge
        if not self.directed:
            if target_id not in self.edges:
                self.edges[target_id] = {}
            self.edges[target_id][source_id] = edge

    def get_edge(self, source_id: NodeType, target_id: NodeType):
        """
        Get an edge from the network

        :param source_id: A tuple ``<agent_category, agent_id>`` stands for source agent.
        :param target_id: A tuple ``<agent_category, agent_id>`` stands for target agent.
        :return: An ``Edge`` object
        """
        return self.edges[source_id][target_id]

    def remove_edge(self, source_id: NodeType, target_id: NodeType):
        """
        Remove an edge from the network

        :param source_id: A tuple ``<agent_category, agent_id>`` stands for source agent.
        :param target_id: A tuple ``<agent_category, agent_id>`` stands for target agent.
        :return: None
        """
        self.edges[source_id].pop(target_id)
        if not self.directed:
            self.edges[target_id].pop(source_id)

    def _get_neighbor_positions(
            self, agent_id: int, category: int
    ) -> List[Tuple[int, int]]:
        neighbor_ids = self.edges[(category, agent_id)]
        if neighbor_ids is None:
            return []
        else:
            return list(neighbor_ids.keys())

    def get_neighbors(self, agent: Agent):
        """
        Get surrounding neighbors of agent.

        :param agent: ``NetworkAgent`` object.
        :return:
        """
        assert hasattr(agent, "category")
        return self._get_neighbor_positions(agent.id, agent.category)

    def _add_agent(self, category: int, agent_id: int):
        agent_tuple: NodeType = (category, agent_id)
        self.nodes.add(agent_tuple)

    def _remove_agent(self, category: int, agent_id: int):
        agent_tuple = (category, agent_id)
        self.nodes.remove(agent_tuple)
        target_edges = self.edges.pop(agent_tuple)
        if not self.directed:
            for target_node, edge in target_edges.items():
                self.edges[target_node].pop(agent_tuple)

    def remove_agent(self, agent: Agent):
        """
        Remove agent from network.

        :param agent: ``NetworkAgent``
        :return: None
        """
        assert hasattr(agent, "category")
        self._remove_agent(agent.category, agent.id)

    def add_agent(self, agent: Agent):
        """
        Add an agent onto the network.

        :param agent: ``NetworkAgent`` object
        :return:
        """
        assert isinstance(agent, NetworkAgent)
        agent.set_category()
        agent._set_network(self)
        self._add_agent(agent.category, agent.id)

    def create_edge(
            self,
            agent_1_id: int,
            category_1: int,
            agent_2_id: int,
            category_2: int,
            **edge_properties,
    ):
        """
        Create a new edge from one agent to another agent.

        :param agent_1_id: ``id`` of source agent
        :param category_1: ``category`` of source agent
        :param agent_2_id: ``id`` of target agent
        :param category_2: ``category`` of target agent
        :param edge_properties: keyword arguments for edge properties.
        :return:
        """
        edge = self.edge_cls(
            category_1, agent_1_id, category_2, agent_2_id, edge_properties
        )
        src_pos = (category_1, agent_1_id)
        dst_pos = (category_2, agent_2_id)
        assert src_pos in self.nodes
        assert dst_pos in self.nodes
        self.add_edge(src_pos, dst_pos, edge)

    def all_agents(self) -> Set[NodeType]:
        """
        Get all agents on the network.

        :return: A list of (`Agent category`, `Agent id`)
        """
        return self.nodes

    def get_node_edges(self, agent: Agent):
        """
        Get the edges from one node.

        :param agent: ``NetworkAgent`` object
        :return: A list of ``Edge`` object
        """
        assert isinstance(agent, NetworkAgent)
        targets = self.edges[(agent.category, agent.id)]
        edges: List[Edge] = []
        for target_node, edge in targets.items():
            edges.append(edge)
        return edges

    def setup_layout_creator(self, layout_creator: "Callable[[nx.Graph], Dict[NodeType, np.ndarray]]"):
        """
        Set the creator function to create network layout.

        The input argument of layout_creator is a networkx graph, and the output was a dict type mapping from node id to
        position.

        For example:
        .. code-block:: python
            import networkx as nx
            n = Network()
            n.setup_layout_creator(lambda G: nx.spring_layout(G))

        :param layout_creator: a callable creating network layout as the code example.
        :return: None
        """
        self._layout_creator = layout_creator

    def update_layout(self):
        """
        Update the layout of network.

        :return: None
        """
        g = nx.DiGraph()
        for start_node in self.edges.keys():
            for end_node in self.edges[start_node].keys():
                g.add_edge(start_node, end_node)
        layout = self._layout_creator(g)
        self.layout = layout

    def get_position(self, agent_category: int, agent_id: int):
        """
        Get the position of agent.

        :param agent_category: The category of agent
        :param agent_id: The id of agent
        :return:
        """
        if (agent_category, agent_id) not in self.layout:
            self.update_layout()
        return self.layout[(agent_category, agent_id)] * 1000

    def setup_agent_connections(
            self,
            agent_lists: List[AgentList],
            network_type: str,
            network_params: dict = None,
    ):
        """
        Set up the connection between agents.
        The name and parameters come from NetworkX package. For example, The documentation for BA scale-free network
        is here: https://networkx.org/documentation/stable/reference/generated/networkx.generators.random_graphs.barabasi_albert_graph.html.
        To create it, network_type is `barabasi_albert_graph`, and parameters should be a dict {"n": 100, "m": 3}.

        :param agent_lists: Initial agent lists containing agents to be placed onto the network.
        :param network_type: str, describing the type of network, which should be the corresponding to networkx.
        :param network_params: A dictionary for parameter values.
        :return: None
        """
        import networkx as nx

        assert isinstance(agent_lists, list)
        node_id = 0
        node_id_to_node_type_map: Dict[int, NodeType] = {}
        for agent_list in agent_lists:
            if len(agent_list) > 0:
                agent_list[0].set_category()
                self.agent_categories[agent_list[0].category] = agent_list
            # category = agent_list[0].category
            # assert isinstance(category, int)
            for agent in agent_list:
                self.add_agent(agent)
                agent._set_network(self)
                node_id_to_node_type_map[node_id] = (agent.category, agent.id)
                node_id += 1

        g = getattr(nx, network_type)(
            len(self.nodes),
            **network_params,
        )

        for edge in g.edges:
            agent_src = node_id_to_node_type_map[edge[0]]
            agent_dest = node_id_to_node_type_map[edge[1]]
            edge_obj = self.edge_cls(
                agent_src[0], agent_src[1], agent_dest[0], agent_dest[1], {}
            )
            self.add_edge(agent_src, agent_dest, edge_obj)
        self._nx_edges = list(g.edges)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import base64
import importlib
import json
import sys
import time
from typing import Dict, List, Tuple, Any, Type, Union, TYPE_CHECKING

import rpyc
import argparse

import cloudpickle

if TYPE_CHECKING:
    from Melodie import Calibrator, Trainer, Scenario, Model

parser = argparse.ArgumentParser()
parser.add_argument("--core_id", help="ID of core")
parser.add_argument("--workdirs", help="Working directories")
parser.add_argument("--role", help="Roles,`calibrator` or `trainer`")
args = parser.parse_args()

workdirs = json.loads(args.workdirs)
sys.path = workdirs + sys.path

from Melodie.global_configs import MelodieGlobalConfig


class ParallelWorker:
    def __init__(self):
        self.role = args.role
        assert self.role in {"trainer", "calibrator"}
        self.core_id = args.core_id
        self.conn = rpyc.connect("localhost", 12233)

    def get_config(self):
        return json.loads(self.conn.root.get_config())

    def get_task(self):
        while 1:
            task_raw = self.conn.root.get_task()
            task = cloudpickle.loads(base64.b64decode(task_raw))
            if task is None:
                time.sleep(0.1)
            else:
                return tuple(task)

    def put_result(self, result):
        return self.conn.root.put_result(result)

    def close(self):
        self.conn.close()

    def run(self):
        config = self.get_config()
        if self.role == "calibrator":
            sub_routine_calibrator(self.core_id, config[0], config[1], self)
        elif self.role == "trainer":
            sub_routine_trainer(self.core_id, config[0], config[1], self)
        else:
            raise NotImplementedError(f"Unrecognized role `{self.role}`")


def get_scenario_manager(config, modules: Dict) -> Tuple[
    Union["Trainer", "Calibrator"], Type["Model"], Type["Scenario"]]:
    from Melodie import Trainer, Calibrator
    classes_dict = {}
    for module_type, content in modules.items():
        class_name, module_name = content
        module = importlib.import_module(module_name)
        cls = getattr(module, class_name)
        classes_dict[module_type] = cls

    trainer: Union[Trainer, Calibrator] = classes_dict["trainer"](
        config=config,
        scenario_cls=classes_dict["scenario"],
        model_cls=classes_dict["model"],
        data_loader_cls=classes_dict["data_loader"],
    )
    trainer.setup()
    trainer.collect_data()
    trainer.subworker_prerun()
    return trainer, classes_dict['scenario'], classes_dict['model']


def sub_routine_trainer(
        proc_id: int,
        modules: Dict[str, Tuple[str, str]],
        config_raw: Dict[str, Any],
        worker: ParallelWorker,
):
    """
    The sub iterator callback for parallelized computing used in Trainer and Calibrator.

    :param proc_id:
    :param modules:
    :param config_raw:
    :return:
    """
    from Melodie import Config, Trainer, Environment, AgentList, Agent
    import logging

    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    logger = logging.getLogger(f"Trainer-processor-{proc_id}")
    logger.info("subroutine started!")
    try:
        config = Config.from_dict(config_raw)
        trainer: Trainer
        trainer, scenario_cls, model_cls = get_scenario_manager(config, modules)
    except BaseException:
        import traceback

        traceback.print_exc()
        return

    while 1:
        try:
            t0 = time.time()

            chrom, d, agent_params = worker.get_task()
            logger.debug(f"processor {proc_id} got chrom {chrom}")
            scenario = scenario_cls()
            scenario.setup()
            scenario.set_params(d)
            model = model_cls(config, scenario)
            scenario.manager = trainer
            model.create()
            model._setup()
            # {category: [{id: 0, param1: 1, param2: 2, ...}]}
            for category, params in agent_params.items():
                agent_container: AgentList[Agent] = getattr(model, category)
                for param in params:
                    agent = agent_container.get_agent(param["id"])
                    agent.set_params(param)
            model.run()
            agent_data = {}
            for container in trainer.container_manager.agent_containers:
                agent_container = getattr(model, container.container_name)
                df = agent_container.to_list(container.recorded_properties)
                agent_data[container.container_name] = df
                for row in df:
                    agent = agent_container.get_agent(row["id"])
                    row["target_function_value"] = trainer.target_function(agent)
                    row["utility"] = trainer.utility(agent)
                    row["agent_id"] = row.pop("id")
            env: Environment = model.environment
            env_data = env.to_dict(trainer.environment_properties)
            dumped = cloudpickle.dumps((chrom, agent_data, env_data))
            t1 = time.time()
            logger.info(
                f"Processor {proc_id}, chromosome {chrom}, time: {MelodieGlobalConfig.Logger.round_elapsed_time(t1 - t0)}s"
            )
            worker.put_result(base64.b64encode(dumped))
        except Exception:
            import traceback

            traceback.print_exc()


def sub_routine_calibrator(
        proc_id: int,
        modules: Dict[str, Tuple[str, str]],
        config_raw: Dict[str, Any],
        worker: ParallelWorker,
):
    """
    The sub iterator callback for parallelized computing used in Trainer and Calibrator.

    :param proc_id:
    :param modules:
    :param config_raw:
    :return:
    """
    from Melodie import Config, Environment
    import logging

    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    logger = logging.getLogger(f"Calibrator-processor-{proc_id}")
    logger.info("subroutine started!")
    try:
        config = Config.from_dict(config_raw)
        calibrator: Calibrator
        calibrator, scenario_cls, model_cls = get_scenario_manager(config, modules)
    except BaseException:
        import traceback

        traceback.print_exc()
        return
    while 1:
        try:
            ret = worker.get_task()
            t0 = time.time()
            chrom, d, env_params = ret
            logger.debug(f"processor {proc_id} got chrom {chrom}")
            scenario: Scenario = scenario_cls()
            scenario.manager = calibrator
            scenario.setup()
            scenario.set_params(d)
            scenario.set_params(env_params)
            model = model_cls(config, scenario)
            model.create()
            model._setup()
            model.run()
            agent_data = {}
            for container_name, props in calibrator.recorded_agent_properties.items():
                agent_container = getattr(model, container_name)
                df = agent_container.to_list(props)
                agent_data[container_name] = df
                for row in df:
                    row["agent_id"] = row.pop("id")
            env: Environment = model.environment
            env_data = env.to_dict(
                calibrator.watched_env_properties
            )
            env_data.update({prop: scenario.to_dict()[prop] for prop in calibrator.properties})
            env_data["target_function_value"] = env_data["distance"] = calibrator.distance(model)
            dumped = cloudpickle.dumps((chrom, agent_data, env_data))
            t1 = time.time()
            logger.info(
                f"Processor {proc_id}, chromosome {chrom}, time: {MelodieGlobalConfig.Logger.round_elapsed_time(t1 - t0)}s"
            )
            worker.put_result(base64.b64encode(dumped))
        except Exception:
            import traceback

            traceback.print_exc()


if __name__ == "__main__":
    worker = ParallelWorker()
    worker.run()



import datetime

from MelodieInfra.jsonobject import StringProperty, BooleanProperty, DateTimeProperty, ListProperty, unicode, JsonObject, \
    DefaultProperty

# https://github.com/dimagi/jsonobject
class User(JsonObject):
    user_name: str = StringProperty(required=True)
    name = StringProperty()
    active = BooleanProperty(default=False)
    date_joined = DateTimeProperty()
    tags = ListProperty(unicode)


user1 = User(
    name='John Doe',
    user_name='jdoe',
    date_joined=datetime.datetime.utcnow(),
    tags=['generic', 'anonymous']
)

print(user1.to_json())
user2 = User({'userName': 'aaaaa'})
print(user2.to_json())



# -*- coding:utf-8 -*-
import os.path
import shutil

import chardet

properties = {
    "_NAME_": "Demostration",
    "_ALIAS_": "Demo",
    "_CREATED_AT_": "",
    "_AUTHOR_": "",
    "_EMAIL_": "",
}

new_project_config_default = {
    "alias": "Demo",
    "name": "Demonstration",
    "path": "",  # create project folder at this folder. for example: if 'path'=c:\users\username\desktop, then
    # all project files will be placed at c:\users\username\desktop\Demonstration
}
_d = os.path.dirname
TEMPLATE_ROOT = os.path.join(_d(__file__), "NewTemplate", "_NAME_")


def create_project(project_config):
    assert os.path.exists(
        project_config["path"]
    ), f'The directory {project_config["path"]} was not found!'
    project_root = os.path.join(project_config["path"], project_config["name"])
    shutil.copytree(TEMPLATE_ROOT, project_root)
    for root, dirs, files in os.walk(project_root):
        for file in files:
            if file.endswith((b".py", b".md", b".txt")):
                file_abs_path = os.path.join(root, file)
                with open(file_abs_path, "rb") as f:
                    bs = f.read()
                print(file, chardet.detect(bs))
                encoding = chardet.detect(bs)["encoding"]
                if encoding is not None:
                    b_formatted = (
                        bs.decode(encoding)
                        .replace("_ALIAS_", project_config["alias"])
                        .encode(encoding)
                    )
                    b_formatted = (
                        b_formatted.decode(encoding)
                        .replace("_NAME_", project_config["name"])
                        .encode(encoding)
                    )
                    with open(file_abs_path, "wb") as f:
                        f.write(b_formatted)


if __name__ == "__main__":
    EXAMPLE_ROOT = os.path.join(_d(_d(_d(__file__))), "examples")
    new_project_config = new_project_config_default.copy()
    new_project_config["path"] = EXAMPLE_ROOT
    create_project(new_project_config)



# -*- coding:utf-8 -*-
from typing import Union, List

import numpy as np

from ._vectorize import apply_float, apply_int, vectorize_int, vectorize_float
from ._vectorize2d import apply2d_float, apply2d_int, vectorize2d_int, vectorize2d_float


def c_vectorize(lst: List[List[Union[int, float]]], property_name: str) -> np.ndarray:
    length = len(lst)
    assert length > 0
    val = getattr(lst[0], property_name)
    if isinstance(val, int):
        return vectorize_int(lst, property_name)
    elif isinstance(val, float):
        return vectorize_float(lst, property_name)
    else:
        raise TypeError


def c_apply(
    lst: List[List[Union[int, float]]], property_name: str, property: np.ndarray
):
    length = len(lst)
    assert length > 0
    val = getattr(lst[0], property_name)
    if isinstance(val, int):
        if property.dtype != np.int64:
            property = property.astype("int64")
        apply_int(lst, property_name, property)
    elif isinstance(val, float):
        if property.dtype != np.float64:
            property = property.astype("float64")
        apply_float(lst, property_name, property)
    else:
        raise TypeError


def c_vectorize_2d(
    list_2d: List[List[Union[int, float]]], property_name: str
) -> np.ndarray:
    ymax = len(list_2d[0])
    xmax = len(list_2d)
    assert xmax, ymax > 0
    val = getattr(list_2d[0][0], property_name)
    if isinstance(val, int):
        return vectorize2d_int(list_2d, property_name)
    elif isinstance(val, float):
        return vectorize2d_float(list_2d, property_name)
    else:
        raise TypeError


def c_apply_2d(
    list_2d: List[List[Union[int, float]]], property_name: str, property: np.ndarray
):
    ymax = len(list_2d[0])
    xmax = len(list_2d)
    assert xmax, ymax > 0
    val = getattr(list_2d[0][0], property_name)
    if isinstance(val, int):
        if property.dtype != np.int64:
            property = property.astype("int64")
        apply2d_int(list_2d, property_name, property)
    elif isinstance(val, float):
        if property.dtype != np.float64:
            property = property.astype("float64")
        apply2d_float(list_2d, property_name, property)
    else:
        raise TypeError


vectorize = c_vectorize
apply = c_apply
vectorize_2d = c_vectorize_2d
apply_2d = c_apply_2d



# cython:language_level=3
# -*- coding:utf-8 -*-
# @Time: 2021/10/3 20:58
# @Author: Zhanyi Hou
# @Email: 1295752786@qq.com
# @File: agent_list.pyx
from libcpp.unordered_map cimport unordered_map as cpp_map
from collections.abc import Sequence
from .basics cimport Agent
cdef class SeqIter:
    """
    The iterator to deal with for-loops in AgentList or other agent containers
    """
    cdef int _i
    cdef list _seq

cdef class DictIter:
    """
    The iterator to deal with for-loops in AgentDict or other agent containers
    """
    cdef object _iter
    cdef dict _dict

cdef class BaseAgentContainer():
    cdef int _id_offset
    cdef object scenario
    cpdef list init_agents(self) except *

    cdef object agent_class
    cdef int initial_agent_num
    cdef object model

cdef class AgentDict(BaseAgentContainer):
    cdef dict agents
    cpdef remove(self, Agent agent)
    cdef _add(self, Agent agent, dict params)
    cpdef Agent get_agent(self, long agent_id)

cdef class AgentList(BaseAgentContainer):
    cdef int _iter_index
    cdef public list agents

    cdef cpp_map[long, long]* indices     
    cdef cpp_map[long, long] _map

    cpdef _add(self, Agent agent, dict params)
    cpdef remove(self, Agent agent)

    # Unordered map is more efficient than cythonized binary search or python dict
    cpdef Agent get_agent(self, long agent_id)

    cdef long _get_index(self, long agent_id) except *
    cdef void _set_index(self, long agent_id, long index) except *

    # cpdef add_agent_id(self, long agent_id, long index)
    cpdef method_foreach(self, str method_name, tuple args) except *
    cpdef list init_agents(self) except *
    cpdef vectorize(self, str prop_name) except *
    cpdef list filter(self, condition) except *
    
cpdef test_container() except *


import copy
import logging
import time
from typing import (
    Dict,
    Tuple,
    Callable,
    Union,
    List,
    Any,
    Optional,
    TYPE_CHECKING,
    Type,
    Iterator,
    cast,
)

import pandas as pd

from MelodieInfra import create_db_conn, Config, MelodieExceptions

from .algorithms import AlgorithmParameters
from .algorithms.ga import MelodieGA
from .data_loader import DataLoader
from .model import Model
from .scenario_manager import Scenario
from .simulator import BaseModellingManager
from .utils.parallel_manager import ParallelManager

if TYPE_CHECKING:
    from .boost.basics import Environment
logger = logging.getLogger(__name__)

pool = None  # on *nix
th_on_thread = None  # for windows


class GACalibratorParams(AlgorithmParameters):
    def __init__(
            self,
            id: int,
            path_num: int,
            generation_num: int,
            strategy_population: int,
            mutation_prob: int,
            strategy_param_code_length: int,
            **kw,
    ):
        super().__init__(id, path_num)

        self.generation_num = generation_num
        self.strategy_population = strategy_population
        self.mutation_prob = mutation_prob
        self.strategy_param_code_length = strategy_param_code_length
        self.parse_params(kw)

    @staticmethod
    def from_dataframe_record(
            record: Dict[str, Union[int, float]]
    ) -> "GACalibratorParams":
        s = GACalibratorParams(
            record["id"],
            record["path_num"],
            record["generation_num"],
            record["strategy_population"],
            record["mutation_prob"],
            record["strategy_param_code_length"],
        )
        s.parse_params(record)
        return s

    def __hash__(self):
        return hash(
            (
                self.id,
                self.path_num,
                self.generation_num,
                self.strategy_population,
                self.mutation_prob,
                self.strategy_param_code_length,
            )
        )


class CalibratorAlgorithmMeta:
    """
    Record the current scenario, params scenario, path and generation
    of trainer.

    """

    def __init__(self):
        self._freeze = False
        self.id_calibrator_scenario = 0
        self.id_calibrator_params_scenario = 1
        self.id_path = 0
        self.id_generation = 0

    def to_dict(self, public_only=False):

        if public_only:
            return {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return copy.copy(self.__dict__)

    def __repr__(self):
        return f"<{self.to_dict()}>"

    def __setattr__(self, key, value):
        if (not hasattr(self, "_freeze")) or (not self._freeze):
            super().__setattr__(key, value)
        else:
            if key in self.__dict__:
                super().__setattr__(key, value)
            else:
                raise MelodieExceptions.General.NoAttributeError(self, key)


class GACalibratorAlgorithmMeta(CalibratorAlgorithmMeta):
    def __init__(self):
        super(GACalibratorAlgorithmMeta, self).__init__()
        self.id_chromosome = 0
        self._freeze = True


class GACalibratorAlgorithm:
    """
    参数：一个tuple
    每次会跑20条染色体，然后将参数缓存起来。
    目标函数从TargetFcnCache中查询。
    """

    def __init__(
            self,
            env_param_names: List[str],
            recorded_env_properties: List[str],
            recorded_agent_properties: Dict[str, List[str]],
            params: GACalibratorParams,
            target_func: "Callable[[Model], Union[float, int]]",
            manager: "Calibrator" = None,
            processors=1,
    ):
        global pool, th_on_thread
        self.manager = manager
        self.params = params
        self.chromosomes = 20
        self.target_func = target_func
        self.env_param_names = env_param_names
        self.recorded_env_properties = recorded_env_properties
        self.recorded_agent_properties = recorded_agent_properties
        lb, ub = self.params.bounds(self.env_param_names)
        self.algorithm: Optional[MelodieGA] = MelodieGA(
            cast("function", self.generate_target_function()),
            len(self.env_param_names),
            self.params.strategy_population,
            self.params.generation_num,
            self.params.mutation_prob,
            lb,
            ub,
            precision=1e-5,
        )
        self.cache: Dict[Tuple[int, int], float] = {}
        # self.agent_container_getters: Dict[str, Callable[[Model], AgentList]] = {}
        # self.agent_ids: Dict[str, List[int]] = {}  # {category : [agent_id]}
        # self.agent_params_defined: Dict[str, List[str]] = {}  # category: [param_name1, param_name2...]
        # self.recorded_agent_properties: Dict[str, List[str]] = {}  # category: [prop1, prop2...]
        # self.recorded_env_properties: List[str] = []  # category: [prop1, prop2...]

        self._chromosome_counter = 0
        self._current_generation = 0
        self.processors = processors
        d = {
            "model": (
                self.manager.model_cls.__name__,
                self.manager.model_cls.__module__,
            ),
            "scenario": (
                self.manager.scenario_cls.__name__,
                self.manager.scenario_cls.__module__,
            ),
            "trainer": (
                self.manager.__class__.__name__,
                self.manager.__class__.__module__,
            ),
            "data_loader": (
                self.manager.df_loader_cls.__name__,
                self.manager.df_loader_cls.__module__,
            ),
        }

        self.parallel_manager = ParallelManager(
            self.processors, configs=(d, self.manager.config.to_dict())
        )
        self.parallel_manager.run("calibrator")

    def stop(self):
        self.parallel_manager.close()

    def get_params(self, id_chromosome: int) -> Dict[str, Any]:
        """
        Pass parameters from the chromosome to the Environment.

        :param id_chromosome:
        :return:
        """
        chromosome_value = self.algorithm.chrom2x(self.algorithm.Chrom)[id_chromosome]
        env_parameters_dict = {}
        for i, param_name in enumerate(self.env_param_names):
            env_parameters_dict[param_name] = chromosome_value[i]
        return env_parameters_dict

    def target_function_to_cache(
            self,
            env_data,
            generation: int,
            id_chromosome: int,
    ):
        """
        Extract the value of target functions from Model, and write them into cache.

        :return:
        """
        self.cache[(generation, id_chromosome)] = env_data["target_function_value"]

    def generate_target_function(self) -> Callable[[], float]:
        """
        Generate the target function.

        :return:
        """

        def f(*args):
            self._chromosome_counter += 1
            value = self.cache[(self._current_generation, self._chromosome_counter)]
            return value

        return f

    def record_agent_properties(
            self,
            agent_data: Dict[str, List[Dict[str, Any]]],
            env_data: Dict[str, Any],
            meta: GACalibratorAlgorithmMeta,
    ):
        """
        Record the property of each agent in the current chromosome.

        :param agent_data: {<agent_container_name>: [{id: 0, prop1: xxx, prop2: xxx, ...}]}
        :param env_data: {env_prop1: xxx, env_prop2: yyy, ...}
        :param meta:
        :return:
        """
        agent_records = {}
        environment_record = {}
        meta_dict = meta.to_dict(public_only=True)

        for container_name, _ in self.recorded_agent_properties.items():
            agent_records[container_name] = []
            data = agent_data[container_name]
            for agent_container_data in data:
                d = {}
                d.update(meta_dict)
                d.update(agent_container_data)

                agent_records[container_name].append(d)
            create_db_conn(self.manager.config).write_dataframe(
                f"{container_name}_calibrator_result",
                pd.DataFrame(agent_records[container_name]),
                if_exists="append",
            )
        environment_record.update(meta_dict)
        environment_record.update(env_data)
        environment_record.pop("target_function_value")

        create_db_conn(self.manager.config).write_dataframe(
            "environment_calibrator_result",
            pd.DataFrame([environment_record]),
            if_exists="append",
        )
        return agent_records, environment_record

    def calc_cov_df(
            self,
            agent_container_df_dict: Dict[str, pd.DataFrame],
            env_df: pd.DataFrame,
            meta,
    ):
        """
        Calculate the coefficient of variation
        :param agent_container_df_dict:
        :param env_df:
        :param meta:
        :return:
        """

        pd.set_option("max_colwidth", 500)
        pd.set_option("display.max_columns", None)
        pd.set_option("display.max_rows", None)
        meta_dict = meta.to_dict(public_only=True)
        meta_dict.pop("id_chromosome")
        for container_name in self.recorded_agent_properties.keys():
            df = agent_container_df_dict[container_name]
            container_agent_record_list = []
            for agent_id in self.recorded_agent_properties[container_name]:
                agent_data = df.loc[df["agent_id"] == agent_id]
                cov_records = {}
                cov_records.update(meta_dict)
                cov_records["agent_id"] = agent_id
                for prop_name in self.recorded_agent_properties[container_name] + [
                    "distance"
                ]:
                    p: pd.Series = agent_data[prop_name]
                    mean = p.mean()
                    cov = p.std() / p.mean()
                    cov_records.update(
                        {prop_name + "_mean": mean, prop_name + "_cov": cov}
                    )
                container_agent_record_list.append(cov_records)
            create_db_conn(self.manager.config).write_dataframe(
                f"{container_name}_calibrator_result_cov",
                pd.DataFrame(container_agent_record_list),
                if_exists="append",
            )
        env_record = {}
        env_record.update(meta_dict)
        for prop_name in (
                self.env_param_names + self.recorded_env_properties + ["distance"]
        ):
            mean = env_df[prop_name].mean()
            cov = env_df[prop_name].std() / env_df[prop_name].mean()
            env_record.update({prop_name + "_mean": mean, prop_name + "_cov": cov})
        create_db_conn(self.manager.config).write_dataframe(
            "environment_calibrator_result_cov",
            pd.DataFrame([env_record]),
            if_exists="append",
        )

    def pre_check(self, meta):
        """
        Check at the beginning of run()
        :return:
        """
        logger.info(f"""Algorithm will run with:
        Meta value: {meta}
        Recording environment parameters: {self.recorded_env_properties}
        Recording Agent agent_lists: {self.recorded_agent_properties}\n""")

    def run(self, scenario: Scenario, meta: Union[GACalibratorAlgorithmMeta]):
        self.pre_check(meta)

        for i in range(self.params.generation_num):
            t0 = time.time()
            self._current_generation = i
            meta.id_generation = i
            logger.info(
                f"======================="
                f"Path {meta.id_path} Generation {i + 1}/{self.params.generation_num}"
                f"======================="
            )

            for id_chromosome in range(self.params.strategy_population):
                params = self.get_params(id_chromosome)
                # params_queue.put(
                #     json.dumps((id_chromosome, scenario.to_json(), params))
                # )
                self.parallel_manager.put_task(
                    (id_chromosome, scenario.to_json(), params)
                )

            agent_records_collector: Dict[str, List[Dict[str, Any]]] = {
                container_name: []
                for container_name in self.recorded_agent_properties.keys()
            }
            env_records_list: List[Dict[str, Any]] = []

            for _id_chromosome in range(self.params.strategy_population):
                chrom, agents_data, env_data = self.parallel_manager.get_result()
                meta.id_chromosome = chrom
                agent_records, env_record = self.record_agent_properties(
                    agents_data, env_data, meta
                )
                for container_name, records in agent_records.items():
                    agent_records_collector[container_name] += records
                env_records_list.append(env_record)
                self.target_function_to_cache(env_data, i, chrom)

            self.calc_cov_df(
                {k: pd.DataFrame(v) for k, v in agent_records_collector.items()},
                pd.DataFrame(env_records_list),
                meta,
            )

            self._chromosome_counter = -1
            self.algorithm.run(1)
            t1 = time.time()
            logger.info("=" * 20 + f"Time Elapsed: {t1 - t0}s" + "=" * 20)


class Calibrator(BaseModellingManager):
    def __init__(
            self,
            config: "Config",
            scenario_cls: "Optional[Type[Scenario]]",
            model_cls: "Optional[Type[Model]]",
            data_loader_cls: Type["DataLoader"],
            processors=1,
    ):
        super().__init__(
            config=config,
            scenario_cls=scenario_cls,
            model_cls=model_cls,
            data_loader_cls=data_loader_cls,
        )
        self.processes = processors
        self.training_strategy: "Optional[Type[SearchingAlgorithm]]" = None
        self.container_name: str = ""

        self.properties: List[str] = []
        self.watched_env_properties: List[str] = []
        self.recorded_agent_properties: Dict[str, List[str]] = {}
        self.algorithm: Optional[GACalibratorAlgorithm] = None
        self.algorithm_instance: Iterator[List[float]] = {}

        self.model: Optional[Model] = None

        self.current_algorithm_meta = GACalibratorAlgorithmMeta()
        self.df_loader_cls = data_loader_cls

    def setup(self):
        """
        Setup method, be sure to inherit this method in custom calibrator class.
        """
        pass

    def collect_data(self):
        """
        Set the agent and environment properties to be collected.

        """
        pass

    def generate_scenarios(self) -> List["Scenario"]:
        """
        Generate scenario objects by the parameter from static tables or scenarios_dataframe.
        
        :return: A list of generated scenarios.
        """
        return self.data_loader.generate_scenarios_from_dataframe(
            "calibrator_scenarios"
        )

    def get_params_scenarios(self) -> List:
        """
        Get the parameters of calibrator parameters from the registered dataframe.

        :return: A list of dict, and each dict contains parameters.
        """
        calibrator_scenarios_table = self.get_dataframe("calibrator_params_scenarios")
        assert isinstance(
            calibrator_scenarios_table, pd.DataFrame
        ), "No learning scenarios table specified!"
        return calibrator_scenarios_table.to_dict(orient="records")

    def run(self):
        """
        The main method for calibrator.

        """
        self.setup()
        self.pre_run()

        scenario_cls = GACalibratorParams
        for scenario in self.scenarios:
            self.current_algorithm_meta.id_calibrator_scenario = scenario.id
            calibration_scenarios = self.get_params_scenarios()
            for calibrator_scenario in calibration_scenarios:
                calibrator_scenario = scenario_cls.from_dataframe_record(
                    calibrator_scenario
                )
                self.current_algorithm_meta.id_calibrator_params_scenario = (
                    calibrator_scenario.id
                )
                for id_trainer_path in range(calibrator_scenario.path_num):
                    self.current_algorithm_meta.id_path = id_trainer_path

                    self.run_once_new(scenario, calibrator_scenario)

    def run_once_new(self, scenario: Scenario, calibration_params: GACalibratorParams):
        """
        Run for one calibration path

        :param scenario: The scenario to run.
        :param calibration_params: calibration parameters.
        :return: None
        """
        self.algorithm = GACalibratorAlgorithm(
            self.properties,
            self.watched_env_properties,
            {},
            calibration_params,
            self.target_function,
            manager=self,
            processors=self.processes,
        )
        self.algorithm.run(scenario, self.current_algorithm_meta)
        self.algorithm.stop()

    def target_function(self, model: "Model") -> Union[float, int]:
        """
        The target function to be minimized

        :param env: Environment of the current model.
        :return:
        """
        return self.distance(model)

    def distance(self, model: "Model") -> float:
        """
        The optimization of calibrator is to minimize the distance.

        Be sure to inherit this function in custom calibrator, and return a float value.

        :param model: The current model after running the current parameter set.
        :return: None
        """
        raise NotImplementedError(
            "Calibrator.distance(model) must be overridden in sub-class!"
        )

    def add_scenario_calibrating_property(self, prop: str):
        """
        Add a property to be tuned in the calibration, and ``prop`` should be a name of property in the environment.

        :param prop: Property name
        :return: None
        """
        assert (
                prop not in self.properties
        ), f'Property "{prop}" is already in the calibrating training_properties!'
        self.properties.append(prop)

    def add_environment_property(self, prop: str):
        """
        Add a property of environment to be recorded in the calibration voyage.

        :param prop: Property name
        :return: None
        """
        assert prop not in self.watched_env_properties
        self.watched_env_properties.append(prop)



from __future__ import absolute_import
from collections import namedtuple, OrderedDict
import copy
import six
import inspect
from .exceptions import (
    DeleteNotAllowed,
    WrappingAttributeError,
)
from .base_properties import JsonProperty, DefaultProperty
from .utils import check_type


JsonObjectClassSettings = namedtuple('JsonObjectClassSettings', ['type_config'])

CLASS_SETTINGS_ATTR = '_$_class_settings'


def get_settings(cls):
    return getattr(cls, CLASS_SETTINGS_ATTR,
                   JsonObjectClassSettings(type_config=TypeConfig()))


def set_settings(cls, settings):
    setattr(cls, CLASS_SETTINGS_ATTR, settings)


class TypeConfig(object):
    """
    This class allows the user to configure dynamic
    type handlers and string conversions for their JsonObject.

    properties is a map from python types to JsonProperty subclasses
    string_conversions is a list or tuple of (regex, python type)-tuples

    This class is used to store the configuration but is not part of the API.
    To configure:

        class Foo(JsonObject):
            # property definitions go here
            # ...

            class Meta(object):
                update_properties = {
                    datetime.datetime: MySpecialDateTimeProperty
                }
                # this is already set by default
                # but you can override with your own modifications
                string_conversions = ((date_re, datetime.date),
                                      (datetime_re, datetime.datetime),
                                      (time_re, datetime.time),
                                      (decimal_re, decimal.Decimal))

    If you now do

        foo = Foo()
        foo.timestamp = datetime.datetime(1988, 7, 7, 11, 8, 0)

    timestamp will be governed by a MySpecialDateTimeProperty
    instead of the default.

    """
    def __init__(self, properties=None, string_conversions=None):
        self._properties = properties if properties is not None else {}

        self._string_conversions = (
            OrderedDict(string_conversions) if string_conversions is not None
            else OrderedDict()
        )
        # cache this
        self.string_conversions = self._get_string_conversions()
        self.properties = self._properties

    def replace(self, properties=None, string_conversions=None):
        return TypeConfig(
            properties=(properties if properties is not None
                        else self._properties),
            string_conversions=(string_conversions if string_conversions is not None
                                else self._string_conversions)
        )

    def updated(self, properties=None, string_conversions=None):
        """
        update properties and string_conversions with the paramenters
        keeping all non-mentioned items the same as before
        returns a new TypeConfig with these changes
        (does not modify original)

        """
        _properties = self._properties.copy()
        _string_conversions = self.string_conversions[:]
        if properties:
            _properties.update(properties)
        if string_conversions:
            _string_conversions.extend(string_conversions)
        return TypeConfig(
            properties=_properties,
            string_conversions=_string_conversions,
        )

    def _get_string_conversions(self):
        result = []
        for pattern, conversion in self._string_conversions.items():
            conversion = (
                conversion if conversion not in self._properties
                else self._properties[conversion](type_config=self).to_python
            )
            result.append((pattern, conversion))
        return result

META_ATTRS = ('properties', 'string_conversions', 'update_properties')


class JsonObjectMeta(type):

    class Meta(object):
        pass

    def __new__(mcs, name, bases, dct):
        cls = super(JsonObjectMeta, mcs).__new__(mcs, name, bases, dct)

        cls.__configure(**{key: value
                           for key, value in cls.Meta.__dict__.items()
                           if key in META_ATTRS})
        cls_settings = get_settings(cls)

        properties = {}
        properties_by_name = {}
        for key, value in dct.items():
            if isinstance(value, JsonProperty):
                properties[key] = value
            elif key.startswith('_'):
                continue
            elif type(value) in cls_settings.type_config.properties:
                property_ = cls_settings.type_config.properties[type(value)](default=value)
                properties[key] = dct[key] = property_
                setattr(cls, key, property_)

        for key, property_ in properties.items():
            property_.init_property(default_name=key,
                                    type_config=cls_settings.type_config)
            assert property_.name is not None, property_
            assert property_.name not in properties_by_name, \
                'You can only have one property named {0}'.format(
                    property_.name)
            properties_by_name[property_.name] = property_

        for base in bases:
            if getattr(base, '_properties_by_attr', None):
                for key, value in base._properties_by_attr.items():
                    if key not in properties:
                        properties[key] = value
                        properties_by_name[value.name] = value

        cls._properties_by_attr = properties
        cls._properties_by_key = properties_by_name
        return cls

    def __configure(cls, properties=None, string_conversions=None,
                    update_properties=None):
        super_settings = get_settings(super(cls, cls))
        assert not properties or not update_properties, \
            "{} {}".format(properties, update_properties)
        type_config = super_settings.type_config
        if update_properties is not None:
            type_config = type_config.updated(properties=update_properties)
        elif properties is not None:
            type_config = type_config.replace(properties=properties)
        if string_conversions is not None:
            type_config = type_config.replace(
                string_conversions=string_conversions)
        set_settings(cls, super_settings._replace(type_config=type_config))
        return cls


class _JsonObjectPrivateInstanceVariables(object):

    def __init__(self, dynamic_properties=None):
        self.dynamic_properties = dynamic_properties or {}


@six.add_metaclass(JsonObjectMeta)
class JsonObjectBase(object):

    _allow_dynamic_properties = True
    _validate_required_lazily = False

    _properties_by_attr = None
    _properties_by_key = None

    _string_conversions = ()

    def __init__(self, _obj=None, **kwargs):
        setattr(self, '_$', _JsonObjectPrivateInstanceVariables())

        self._obj = check_type(_obj, dict,
                               'JsonObject must wrap a dict or None')
        self._wrapped = {}

        for key, value in list(self._obj.items()):
            try:
                self.set_raw_value(key, value)
            except AttributeError:
                raise WrappingAttributeError(
                    "can't set attribute corresponding to {key!r} "
                    "on a {cls} while wrapping {data!r}".format(
                        cls=self.__class__,
                        key=key,
                        data=_obj,
                    )
                )

        for attr, value in kwargs.items():
            try:
                setattr(self, attr, value)
            except AttributeError:
                raise WrappingAttributeError(
                    "can't set attribute {key!r} "
                    "on a {cls} while wrapping {data!r}".format(
                        cls=self.__class__,
                        key=attr,
                        data=_obj,
                    )
                )

        for key, value in self._properties_by_key.items():
            if key not in self._obj:
                try:
                    d = value.default()
                except TypeError:
                    d = value.default(self)
                self[key] = d

    def set_raw_value(self, key, value):
        wrapped = self.__wrap(key, value)
        if key in self._properties_by_key:
            self[key] = wrapped
        else:
            setattr(self, key, wrapped)

    @classmethod
    def properties(cls):
        return cls._properties_by_attr.copy()

    @property
    def __dynamic_properties(self):
        return getattr(self, '_$').dynamic_properties

    @classmethod
    def wrap(cls, obj):
        self = cls(obj)
        return self

    def validate(self, required=True):
        for key, value in self._wrapped.items():
            self.__get_property(key).validate(value, required=required)

    def to_json(self):
        self.validate()
        return copy.deepcopy(self._obj)

    def __get_property(self, key):
        try:
            return self._properties_by_key[key]
        except KeyError:
            return DefaultProperty(type_config=get_settings(self).type_config)

    def __wrap(self, key, value):
        property_ = self.__get_property(key)

        if value is None:
            return None

        return property_.wrap(value)

    def __unwrap(self, key, value):
        property_ = self.__get_property(key)
        if value is None:
            wrapped, unwrapped = None, None
        else:
            wrapped, unwrapped = property_.unwrap(value)

        if isinstance(wrapped, JsonObjectBase):
            # validate containers but not objects
            recursive_kwargs = {'recursive': False}
        else:
            # omit the argument for backwards compatibility of custom properties
            # that do not contain `recursive` in their signature
            # and let the default of True shine through
            recursive_kwargs = {}
        property_.validate(
            wrapped,
            required=not self._validate_required_lazily,
            **recursive_kwargs,
        )
        return wrapped, unwrapped

    def __setitem__(self, key, value):
        wrapped, unwrapped = self.__unwrap(key, value)
        self._wrapped[key] = wrapped
        if self.__get_property(key).exclude(unwrapped):
            self._obj.pop(key, None)
        else:
            self._obj[key] = unwrapped
        if key not in self._properties_by_key:
            assert key not in self._properties_by_attr
            self.__dynamic_properties[key] = wrapped
            super(JsonObjectBase, self).__setattr__(key, wrapped)

    def __is_dynamic_property(self, name):
        return (
            name not in self._properties_by_attr and
            not name.startswith('_') and
            not inspect.isdatadescriptor(getattr(self.__class__, name, None))
        )

    def __setattr__(self, name, value):
        if self.__is_dynamic_property(name):
            if self._allow_dynamic_properties:
                self[name] = value
            else:
                raise AttributeError(
                    "{0!r} is not defined in schema "
                    "(not a valid property)".format(name)
                )
        else:
            super(JsonObjectBase, self).__setattr__(name, value)

    def __delitem__(self, key):
        if key in self._properties_by_key:
            raise DeleteNotAllowed(key)
        else:
            if not self.__is_dynamic_property(key):
                raise KeyError(key)
            del self._obj[key]
            del self._wrapped[key]
            del self.__dynamic_properties[key]
            super(JsonObjectBase, self).__delattr__(key)

    def __delattr__(self, name):
        if name in self._properties_by_attr:
            raise DeleteNotAllowed(name)
        elif self.__is_dynamic_property(name):
            del self[name]
        else:
            super(JsonObjectBase, self).__delattr__(name)

    def __repr__(self):
        name = self.__class__.__name__
        predefined_properties = self._properties_by_attr.keys()
        predefined_property_keys = set(self._properties_by_attr[p].name
                                       for p in predefined_properties)
        dynamic_properties = (set(self._wrapped.keys())
                              - predefined_property_keys)
        properties = sorted(predefined_properties) + sorted(dynamic_properties)
        return u'{name}({keyword_args})'.format(
            name=name,
            keyword_args=', '.join('{key}={value!r}'.format(
                key=key,
                value=getattr(self, key)
            ) for key in properties),
        )


class _LimitedDictInterfaceMixin(object):
    """
    mindlessly farms selected dict methods out to an internal dict

    really only a separate class from JsonObject
    to keep this mindlessness separate from the methods
    that need to be more carefully understood

    """
    _wrapped = None

    def keys(self):
        return self._wrapped.keys()

    def items(self):
        return self._wrapped.items()

    def iteritems(self):
        return self._wrapped.iteritems()

    def __contains__(self, item):
        return item in self._wrapped

    def __getitem__(self, item):
        return self._wrapped[item]

    def __iter__(self):
        return iter(self._wrapped)

    def __len__(self):
        return len(self._wrapped)


def get_dynamic_properties(obj):
    return getattr(obj, '_$').dynamic_properties.copy()



import cProfile
import io
import pstats


def run_profile(callback):
    """
    Run a function with a profiler.
    When the callback finished or pressed Ctrl+C(KeyBoardInterrupt), the profile will be printed at the screen.
    :param callback:
    :return:
    """
    pr = cProfile.Profile()
    pr.enable()
    try:
        callback()
    except KeyboardInterrupt:
        pass
    pr.disable()
    s = io.StringIO()
    sortby = pstats.SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print(s.getvalue())



# cython:language_level=3
from typing import Dict, Any
from typing import Tuple, List, Dict, Optional, TYPE_CHECKING

import pandas as pd

from Melodie.basic import MelodieExceptions

if TYPE_CHECKING:
    from Melodie import Model, Scenario

cdef class Element:
    cpdef void set_params(self, dict params) except *

cdef class Agent(Element):
    cdef public long id
    cdef public object model, scenario
    cpdef dict to_dict(self, list properties) except *


cdef class Environment(Element):
    cdef public object model
    cdef public object scenario

    cpdef dict to_dict(self, list properties)

    cpdef to_dataframe(self, properties: List[str])
